import asyncio
import site
import socket
import subprocess

from fhempy.lib.generic import FhemModule

from .. import fhem, utils


class esphome(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.proc = None
        self._set_list = {"start": {}, "stop": {}, "restart": {}}
        self.set_set_config(self._set_list)
        self._attr_list = {"disable": {"default": "0", "options": "0,1"}}
        self.set_attr_config(self._attr_list)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        if self._attr_disable == "1":
            return

        await self.start_process()

        if await fhem.AttrVal(self.hash["NAME"], "room", "") == "":
            await fhem.CommandAttr(self.hash, hash["NAME"] + " room ESPHome")
            self.create_async_task(self.create_weblink())

    async def start_process(self):
        self._esphomeargs = [
            site.getuserbase() + "/bin/esphome",
            "esphome_config/",
            "dashboard",
        ]

        try:
            self.proc = subprocess.Popen(self._esphomeargs)
        except Exception:
            try:
                self._esphomeargs = ["esphome", "dashboard", "esphome_config/"]
                self.proc = subprocess.Popen(self._esphomeargs)
            except Exception:
                return "Failed to execute esphome"

        await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)

    async def stop_process(self):
        if self.proc:
            self.proc.terminate()
            self.proc = None
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def create_weblink(self):
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        await fhem.CommandDefine(
            self.hash, "esphome_dashboard weblink iframe http://" + local_ip + ":6052/"
        )
        await fhem.CommandAttr(
            self.hash,
            "esphome_dashboard htmlattr width='900' height='700' frameborder='0' marginheight='0' marginwidth='0'",
        )
        await fhem.CommandAttr(self.hash, "esphome_dashboard room ESPHome")

    # FHEM FUNCTION
    async def Undefine(self, hash):
        if self.proc:
            self.proc.terminate()
        return await super().Undefine(hash)

    async def set_attr_disable(self, hash):
        if self._attr_disable == "0":
            await self.start_process()
        else:
            await self.stop_process()

    async def set_start(self, hash, params):
        await self.stop_process()
        await self.start_process()
        return ""

    async def set_stop(self, hash, params):
        await self.stop_process()
        return ""

    async def set_restart(self, hash, params):
        return await self.set_start(hash, None)



import logging
from dataclasses import dataclass, field
from typing import Dict, Any, List, Callable

from . import bluetooth, zigbee
from .utils import DOMAIN

_LOGGER = logging.getLogger(__name__)


# TODO: rewrite all usage to dataclass
@dataclass
class XiaomiDevice:
    did: str  # unique Xiaomi did
    model: str  # Xiaomi model
    mac: str
    type: str  # gateway, zigbee, ble, mesh
    online: bool

    lumi_spec: list
    miot_spec: list

    device_info: Dict[str, Any]

    stats: "XiaomiEntity"

    extra: Dict[str, Any] = field(default_factory=dict)

    # all device entities except stats
    entities: Dict[str, "XiaomiEntity"] = field(default_factory=dict)

    gateways: List["Gateway3"] = field(default_factory=list)


class DevicesRegistry:
    """Global registry for all gateway devices. Because BLE devices updates
    from all gateway simultaniosly.

    Key - device did, `numb` for wifi and mesh devices, `lumi.ieee` for zigbee
    devices, `blt.3.alphanum` for ble devices, `group.numb` for mesh groups.
    """

    devices: Dict[str, dict] = {}
    setups: Dict[str, Callable] = None

    defaults: Dict[str, dict] = {}

    def add_setup(self, domain: str, handler):
        """Add hass device setup funcion."""
        self.setups[domain] = handler

    def add_entity(self, domain: str, device: dict, attr: str):
        if self not in device["gateways"]:
            device["gateways"].append(self)

        if domain is None or attr in device["entities"]:
            return

        # instant add entity to prevent double setup
        device["entities"]["main"] = None

        self.setups[domain](self, device, attr)

    def set_entity(self, fhem_dev):
        fhem_dev.device["entities"]["main"] = fhem_dev

    def remove_entity(self, fhem_dev):
        fhem_dev.device["entities"].pop("main")

    def add_stats(self, device: dict):
        if "stats" in device:
            return

        device["stats"] = None

        self.setups["sensor"](self, device, device["type"])

    def set_stats(self, fhem_dev):
        fhem_dev.device["stats"] = fhem_dev

    def remove_stats(self, fhem_dev):
        fhem_dev.device.pop("stats")

    def find_or_create_device(self, device: dict) -> dict:
        type_ = device["type"]
        did = device["did"] if type_ != "ble" else device["mac"].lower()
        if did in self.devices:
            return self.devices[did]

        self.devices[did] = device

        # update device with specs
        if type_ in ("gateway", "zigbee"):
            device.update(zigbee.get_device(device["model"]))
        elif type_ == "mesh":
            device.update(bluetooth.get_device(device["model"], "Mesh"))
        # elif type_ == 'ble':
        #     device.update(bluetooth.get_device(device['model'], 'BLE'))

        model = device["model"]
        if model in self.defaults:
            device.update(self.defaults[model])

        if did in self.defaults:
            device.update(self.defaults[did])

        mac = device["mac"].lower()
        if did != mac and mac in self.defaults:
            device.update(self.defaults[mac])

        device["entities"] = {}
        device["gateways"] = []

        return device


class XiaomiEntity:
    _ignore_offline = None
    _state = None

    def __init__(self, gateway: "Gateway3", device: dict, attr: str):
        self.gw = gateway
        self.device = device

        self.attr = attr
        self._attrs = {}

        self._unique_id = f"{device.get('entity_name', device['mac'])}_{attr}"
        self._name = device["device_name"] + " " + attr.replace("_", " ").title()

        self.entity_id = f"{DOMAIN}.{self._unique_id}"

    def debug(self, message: str):
        self.gw.debug(f"{self.entity_id} | {message}")

    # @property
    # def entity_registry_enabled_default(self):
    #     return False

    @property
    def should_poll(self) -> bool:
        return False

    @property
    def unique_id(self):
        return self._unique_id

    @property
    def name(self):
        return self._name

    @property
    def available(self) -> bool:
        gw_available = any(gateway.available for gateway in self.device["gateways"])
        return gw_available and (
            self.device.get("online", True) or self._ignore_offline
        )

    @property
    def device_state_attributes(self):
        return self._attrs

    @property
    def device_info(self):
        """
        https://developers.home-assistant.io/docs/device_registry_index/
        """
        type_ = self.device["type"]
        if type_ == "gateway":
            return {
                "connections": {("mac", self.device["wlan_mac"])},
                "identifiers": {(DOMAIN, self.device["mac"])},
                "manufacturer": self.device["device_manufacturer"],
                "model": self.device["device_model"],
                "name": self.device["device_name"],
                "sw_version": self.device["fw_ver"],
            }
        elif type_ == "zigbee":
            return {
                "connections": {(type_, self.device["mac"])},
                "identifiers": {(DOMAIN, self.device["mac"])},
                "manufacturer": self.device.get("device_manufacturer"),
                "model": self.device["device_model"],
                "name": self.device["device_name"],
                "sw_version": self.device.get("fw_ver"),
                "via_device": (DOMAIN, self.gw.device["mac"]),
            }
        else:  # ble and mesh
            return {
                "connections": {("bluetooth", self.device["mac"])},
                "identifiers": {(DOMAIN, self.device["mac"])},
                "manufacturer": self.device.get("device_manufacturer"),
                "model": self.device["device_model"],
                "name": self.device["device_name"],
                "via_device": (DOMAIN, self.gw.device["mac"]),
            }

    def update(self, data: dict):
        pass



# lot of parts copied from HomeAssistant, many thanks!

import asyncio
import concurrent
import functools
import inspect
import json
import logging
import os
import sys
from pathlib import Path
from subprocess import PIPE, Popen

import pkg_resources

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

pip_lock = asyncio.Lock()

if sys.version_info[:2] >= (3, 8):
    from importlib.metadata import (  # pylint: disable=no-name-in-module,import-error
        PackageNotFoundError,
        version,
    )
else:
    from importlib_metadata import (  # pylint: disable=import-error
        PackageNotFoundError,
        version,
    )


def is_virtual_env() -> bool:
    """Return if we run in a virtual environment."""
    # Check supports venv && virtualenv
    return getattr(sys, "base_prefix", sys.prefix) != sys.prefix or hasattr(
        sys, "real_prefix"
    )


def is_docker_env() -> bool:
    """Return True if we run in a docker env."""
    return Path("/.dockerenv").exists()


def pip_kwargs(config_dir):
    """Return keyword arguments for PIP install."""
    is_docker = is_docker_env()
    kwargs = {
        # "constraints": os.path.join(os.path.dirname(__file__), CONSTRAINT_FILE),
        "no_cache_dir": is_docker,
    }
    if "WHEELS_LINKS" in os.environ:
        kwargs["find_links"] = os.environ["WHEELS_LINKS"]
    if not (config_dir is None or is_virtual_env()) and not is_docker:
        kwargs["target"] = os.path.join(config_dir, "deps")
    return kwargs


def check_dependencies(module):
    """Checks the manifest of a specific module and check installation
    of dependencies
    """
    try:
        from fhempy import lib

        initfile = inspect.getfile(lib)
        fhempy_root = os.path.dirname(initfile)
        with open(fhempy_root + "/" + module + "/manifest.json", "r") as f:
            manifest = json.load(f)

            if "requirements" in manifest:
                for req in manifest["requirements"]:
                    logger.debug("Check requirement: " + req)
                    if is_installed(req) is False:
                        logger.debug("  NOK")
                        return False
                    else:
                        logger.debug("  OK")
    except FileNotFoundError:
        logger.error("manifest.json not found!")

    return True


async def force_update_package(package):
    kwargs = pip_kwargs(None)
    ret = False
    async with pip_lock:
        with concurrent.futures.ThreadPoolExecutor() as pool:
            ret = await asyncio.get_event_loop().run_in_executor(
                pool, functools.partial(install_package, package, **kwargs)
            )
    return ret


async def check_and_install_dependencies(module):
    """Checks the manifest of a specific module and starts installation
    of dependencies
    """
    try:
        async with pip_lock:
            kwargs = pip_kwargs(None)
            from fhempy import lib

            initfile = inspect.getfile(lib)
            fhempy_root = os.path.dirname(initfile)
            with open(fhempy_root + "/" + module + "/manifest.json", "r") as f:
                manifest = json.load(f)

                if "requirements" in manifest:
                    for req in manifest["requirements"]:
                        if is_installed(req) is False:
                            inst_tries = 0
                            while inst_tries < 3:
                                with concurrent.futures.ThreadPoolExecutor() as pool:
                                    ret = (
                                        await asyncio.get_event_loop().run_in_executor(
                                            pool,
                                            functools.partial(
                                                install_package, req, **kwargs
                                            ),
                                        )
                                    )
                                if ret:
                                    break
                                inst_tries += 1
    except FileNotFoundError:
        pass

    return


def is_installed(package: str) -> bool:
    """Check if a package is installed and will be loaded when we import it.
    Returns True when the requirement is met.
    Returns False when the package is not installed or doesn't meet req.
    """
    try:
        req = pkg_resources.Requirement.parse(package)
    except ValueError:
        raise

    try:
        ret = version(req.project_name) in req
        return ret
    except PackageNotFoundError:
        return False


def install_package(
    package: str,
    upgrade: bool = True,
    target: [str] = None,
    constraints: [str] = None,
    find_links: [str] = None,
    no_cache_dir: [bool] = False,
) -> bool:
    """Install a package on PyPi. Accepts pip compatible package strings.
    Return boolean if install successful.
    """
    # Not using 'import pip; pip.main([])' because it breaks the logger
    logger.info("Attempting install of %s", package)
    env = os.environ.copy()
    args = [sys.executable, "-m", "pip", "install", "--quiet", package]
    if no_cache_dir:
        args.append("--no-cache-dir")
    if upgrade:
        args.append("--upgrade")
    if constraints is not None:
        args += ["--constraint", constraints]
    if find_links is not None:
        args += ["--find-links", find_links, "--prefer-binary"]
    if target:
        assert not is_virtual_env()
        # This only works if not running in venv
        args += ["--user"]
        env["PYTHONUSERBASE"] = os.path.abspath(target)
        if sys.platform != "win32":
            # Workaround for incompatible prefix setting
            # See http://stackoverflow.com/a/4495175
            args += ["--prefix="]
    process = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=env)
    _, stderr = process.communicate()
    if process.returncode != 0:
        logger.error(
            "Unable to install package %s: %s",
            package,
            stderr.decode("utf-8").lstrip().strip(),
        )
        return False

    logger.info("Successfully installed " + package + " update!")
    return True



from datetime import datetime
from typing import Optional

# Bluetooth Model: [Manufacturer, Device Name, Device Model]
# params: [siid, piid, hass attr name, hass domain]
DEVICES = [
    {
        # MiBeacon from official support
        152: ["Xiaomi", "Flower Care", "HHCCJCY01"],
        349: ["Xiaomi", "Flower Pot", "HHCCPOT002"],
        426: ["Xiaomi", "TH Sensor", "LYWSDCGQ/01ZM"],
        794: ["Xiaomi", "Door Lock", "MJZNMS02LM"],
        839: ["Xiaomi", "Qingping TH Sensor", "CGG1"],
        903: ["Xiaomi", "ZenMeasure TH", "MHO-C401"],
        982: ["Xiaomi", "Qingping Door Sensor", "CGH1"],
        1034: ["Xiaomi", "Mosquito Repellent", "WX08ZM"],
        1115: ["Xiaomi", "TH Clock", "LYWSD02MMC"],
        1161: ["Xiaomi", "Toothbrush T500", "MES601"],
        1249: ["Xiaomi", "Magic Cube", "XMMF01JQD"],
        1371: ["Xiaomi", "TH Sensor 2", "LYWSD03MMC"],
        1398: ["Xiaomi", "Alarm Clock", "CGD1"],
        1433: ["Xiaomi", "Door Lock", "MJZNMS03LM"],
        1647: ["Xiaomi", "Qingping TH Lite", "CGDK2"],
        1694: ["Aqara", "Door Lock N100", "ZNMS16LM"],
        1695: ["Aqara", "Door Lock N200", "ZNMS17LM"],
        1747: ["Xiaomi", "ZenMeasure Clock", "MHO-C303"],
        1983: ["Yeelight", "Button S1", "YLAI003"],
        2038: ["Xiaomi", "Night Light 2", "MJYD02YL-A"],  # 15,4103,4106,4119,4120
        2147: ["Xiaomi", "Water Leak Sensor", "SJWS01LM"],
        2443: ["Xiaomi", "Door Sensor 2", "MCCGQ02HL"],
        2444: ["Xiaomi", "Door Lock", "XMZNMST02YD"],
        2455: ["Honeywell", "Smoke Alarm", "JTYJ-GD-03MI"],
        2480: ["Xiaomi", "Safe Box", "BGX-5/X1-3001"],
        2691: ["Xiaomi", "Qingping Motion Sensor", "CGPR1"],
        # logs: https://github.com/AlexxIT/XiaomiGateway3/issues/180
        2701: ["Xiaomi", "Motion Sensor 2", "RTCGQ02LM"],  # 15,4119,4120
        2888: ["Xiaomi", "Qingping TH Sensor", "CGG1"],  # same model as 839?!
    },
    {
        # Mesh Light
        0: ["Xiaomi", "Mesh Group", "Mesh Group"],
        948: ["Yeelight", "Mesh Downlight", "YLSD01YL"],
        995: ["Yeelight", "Mesh Bulb E14", "YLDP09YL"],
        996: ["Yeelight", "Mesh Bulb E27", "YLDP10YL"],
        997: ["Yeelight", "Mesh Spotlight", "YLSD04YL"],
        1771: ["Xiaomi", "Mesh Bulb", "MJDP09YL"],
        1772: ["Xiaomi", "Mesh Downlight", "MJTS01YL/MJTS003"],
        2076: ["Yeelight", "Mesh Downlight M2", "YLTS02YL/YLTS04YL"],
        2293: ["Unknown", "Mesh Lightstrip (RF ready)", "crzm.light.wy0a01"],
        2342: ["Yeelight", "Mesh Bulb M2", "YLDP25YL/YLDP26YL"],
        2351: ["Unknown", "Mesh Downlight", "lemesh.light.wy0c05"],
        2584: ["XinGuang", "XinGuang Smart Light", "LIBMDA09X"],
        3164: ["Unknown", "Mesh Downlight (RF ready)", "lemesh.light.wy0c07"],
        3416: ["Unknown", "Mesh Downlight", "090615.light.mlig01"],
        3531: ["Unknown", "ightctl Light", "lemesh.light.wy0c08"],
        "miot_spec": [
            [2, 1, "light", "light"],
            [2, 2, "brightness", None],
            [2, 3, "color_temp", None],
        ],
    },
    {
        # Mesh Switches
        1946: ["Xiaomi", "Mesh Double Wall Switch", "DHKG02ZM"],
        "miot_spec": [
            [2, 1, "left_switch", "switch"],
            [3, 1, "right_switch", "switch"],
        ],
    },
    {
        1945: ["Xiaomi", "Mesh Wall Switch", "DHKG01ZM"],
        2007: ["Unknown", "Mesh Switch Controller", "lemesh.switch.sw0a01"],
        3150: ["XinGuang", "Mesh Switch", "wainft.switch.sw0a01"],
        "miot_spec": [[2, 1, "switch", "switch"]],
    },
    {
        2093: ["PTX", "Mesh Triple Wall Switch", "PTX-TK3/M"],
        3878: ["PTX", "Mesh Triple Wall Switch", "PTX-SK3M"],
        "miot_spec": [
            [2, 1, "left_switch", "switch"],
            [3, 1, "middle_switch", "switch"],
            [4, 1, "right_switch", "switch"],
            [8, 1, "backlight", "switch"],
            [8, 2, "left_smart", "switch"],
            [8, 3, "middle_smart", "switch"],
            [8, 4, "right_smart", "switch"],
        ],
    },
    {
        2257: ["PTX", "Mesh Double Wall Switch", "PTX-SK2M"],
        "miot_spec": [
            [2, 1, "left_switch", "switch"],
            [3, 1, "right_switch", "switch"],
            [8, 1, "backlight", "switch"],
            [8, 2, "left_smart", "switch"],
            [8, 3, "right_smart", "switch"],
        ],
    },
    {
        2258: ["PTX", "Mesh Single Wall Switch", "PTX-SK1M"],
        "miot_spec": [
            [2, 1, "switch", "switch"],
            [8, 1, "backlight", "switch"],
            [8, 2, "smart", "switch"],
        ],
    },
    {
        2717: ["Xiaomi", "Mesh Triple Wall Switch", "ZNKG03HL/ISA-KG03HL"],
        "miot_spec": [
            [2, 1, "left_switch", "switch"],
            [3, 1, "middle_switch", "switch"],
            [4, 1, "right_switch", "switch"],
            [6, 1, "humidity", "sensor"],
            [6, 7, "temperature", "sensor"],
        ],
    },
    {
        3083: ["Xiaomi", "Mi Smart Electrical Outlet", "ZNCZ01ZM"],
        "miot_spec": [
            [2, 1, "outlet", "switch"],
            [3, 1, "power", "sensor"],
            [4, 1, "backlight", "switch"],
        ],
    },
    {
        2715: ["Xiaomi", "Mesh Single Wall Switch", "ZNKG01HL"],
        "miot_spec": [
            [2, 1, "switch", "switch"],
            [6, 1, "humidity", "sensor"],
            [6, 7, "temperature", "sensor"],
        ],
    },
    {
        2716: ["Xiaomi", "Mesh Double Wall Switch", "ZNKG02HL"],
        "miot_spec": [
            [2, 1, "left_switch", "switch"],
            [3, 1, "right_switch", "switch"],
            [6, 1, "humidity", "sensor"],
            [6, 7, "temperature", "sensor"],
        ],
    },
]

# if color temp not default 2700..6500
COLOR_TEMP = {
    # https://github.com/AlexxIT/XiaomiGateway3/issues/350
    2351: [3100, 3100],
    2584: [3000, 6400],
    3531: [3000, 6400],
}
# if max brightness not default 65535
MAX_BRIGHTNESS = {
    2293: 100,
    2351: 100,
    2584: 100,
    3164: 100,
    3416: 100,
    3531: 100,
}

BLE_FINGERPRINT_ACTION = [
    "Match successful",
    "Match failed",
    "Timeout",
    "Low quality",
    "Insufficient area",
    "Skin is too dry",
    "Skin is too wet",
]
BLE_DOOR_ACTION = [
    "Door is open",
    "Door is closed",
    "Timeout is not closed",
    "Knock on the door",
    "Breaking the door",
    "Door is stuck",
]
BLE_LOCK_ACTION = {
    0b0000: "Unlock outside the door",
    0b0001: "Lock",
    0b0010: "Turn on anti-lock",
    0b0011: "Turn off anti-lock",
    0b0100: "Unlock inside the door",
    0b0101: "Lock inside the door",
    0b0110: "Turn on child lock",
    0b0111: "Turn off child lock",
    0b1111: None,
}
BLE_LOCK_METHOD = {
    0b0000: "bluetooth",
    0b0001: "password",
    0b0010: "biological",
    0b0011: "key",
    0b0100: "turntable",
    0b0101: "nfc",
    0b0110: "one-time password",
    0b0111: "two-step verification",
    0b1000: "coercion",
    0b1010: "manual",
    0b1011: "automatic",
    0b1111: None,
}
BLE_LOCK_ERROR = {
    0xC0DE0000: "Frequent unlocking with incorrect password",
    0xC0DE0001: "Frequent unlocking with wrong fingerprints",
    0xC0DE0002: "Operation timeout (password input timeout)",
    0xC0DE0003: "Lock picking",
    0xC0DE0004: "Reset button is pressed",
    0xC0DE0005: "The wrong key is frequently unlocked",
    0xC0DE0006: "Foreign body in the keyhole",
    0xC0DE0007: "The key has not been taken out",
    0xC0DE0008: "Error NFC frequently unlocks",
    0xC0DE0009: "Timeout is not locked as required",
    0xC0DE000A: "Failure to unlock frequently in multiple ways",
    0xC0DE000B: "Unlocking the face frequently fails",
    0xC0DE000C: "Failure to unlock the vein frequently",
    0xC0DE000D: "Hijacking alarm",
    0xC0DE000E: "Unlock inside the door after arming",
    0xC0DE000F: "Palmprints frequently fail to unlock",
    0xC0DE0010: "The safe was moved",
    0xC0DE1000: "The battery level is less than 10%",
    0xC0DE1001: "The battery is less than 5%",
    0xC0DE1002: "The fingerprint sensor is abnormal",
    0xC0DE1003: "The accessory battery is low",
    0xC0DE1004: "Mechanical failure",
}

ACTIONS = {
    1249: {0: "right", 1: "left"},
    1983: {0: "single", 0x010000: "double", 0x020000: "hold"},
    2147: {0: "single"},
}


def get_ble_domain(param: str, value) -> Optional[str]:
    if param in (
        "contact",
        "gas",
        "light",
        "lock",
        "motion",
        "opening",
        "power",
        "sleep",
        "smoke",
        "water_leak",
    ):
        return "binary_sensor"

    if param in (
        "battery",
        "conductivity",
        "formaldehyde",
        "humidity",
        "idle_time",
        "illuminance",
        "moisture",
        "supply",
        "temperature",
    ):
        return "sensor"

    if param == "action" and value not in ("motion", "tracker", "weigth"):
        return "sensor"

    if param == "tracker":
        return "device_tracker"

    return None


def parse_xiaomi_ble(event: dict, pdid: int) -> Optional[dict]:
    """Parse Xiaomi BLE Data
    https://iot.mi.com/new/doc/embedded-development/ble/object-definition
    """
    eid = event["eid"]
    data = bytes.fromhex(event["edata"])
    length = len(data)

    if eid == 0x1001 and length == 3:  # 4097
        value = int.from_bytes(data, "little")
        return {
            "action": ACTIONS[pdid][value]
            if pdid in ACTIONS and value in ACTIONS[pdid]
            else value
        }

    elif eid == 0x1002 and length == 1:  # 4098
        # No sleep (0x00), falling asleep (0x01)
        return {"sleep": data[0]}  # 1 => true

    elif eid == 0x1003 and length == 1:  # 4099
        # Signal strength value
        return {"rssi": data[0]}

    elif eid == 0x1004 and length == 2:  # 4100
        return {"temperature": int.from_bytes(data, "little", signed=True) / 10.0}

    elif eid == 0x1005 and length == 2:  # 4101
        # Kettle, thanks https://github.com/custom-components/ble_monitor/
        return {"power": data[0], "temperature": data[1]}

    elif eid == 0x1006 and length == 2:  # 4102
        # Humidity percentage, ranging from 0-1000
        value = int.from_bytes(data, "little") / 10.0
        if pdid in (903, 1371):
            # two models has bug, they increase humidity on each data by 0.1
            value = int(value)
        return {"humidity": value}

    elif eid == 0x1007 and length == 3:  # 4103
        value = int.from_bytes(data, "little")

        if pdid == 2038:
            # Night Light 2: 1 - no light, 100 - light
            return {"light": int(value >= 100)}

        # Range: 0-120000, lux
        return {"illuminance": value}

    elif eid == 0x1008 and length == 1:  # 4104
        # Humidity percentage, range: 0-100
        return {"moisture": data[0]}

    elif eid == 0x1009 and length == 2:  # 4105
        # Soil EC value, Unit us/cm, range: 0-5000
        return {"conductivity": int.from_bytes(data, "little")}

    elif eid == 0x100A:  # 4106
        # TODO: lock timestamp
        return {"battery": data[0]}

    elif eid == 0x100D and length == 4:  # 4109
        return {
            "temperature": int.from_bytes(data[:2], "little", signed=True) / 10.0,
            "humidity": int.from_bytes(data[2:], "little") / 10.0,
        }

    elif eid == 0x100E and length == 1:  # 4110
        # 1 => true => on => unlocked
        # 0x00: unlock state (all bolts retracted)
        # TODO: other values
        return {"lock": 1 if data[0] == 0 else 0}

    elif eid == 0x100F and length == 1:  # 4111
        # 1 => true => on => dooor opening
        return {"opening": 1 if data[0] == 0 else 0}

    elif eid == 0x1010 and length == 2:  # 4112
        return {"formaldehyde": int.from_bytes(data, "little") / 100.0}

    elif eid == 0x1012 and length == 1:  # 4114
        return {"opening": data[0]}  # 1 => true => open

    elif eid == 0x1013 and length == 1:  # 4115
        # Remaining percentage, range 0~100
        return {"supply": data[0]}

    elif eid == 0x1014 and length == 1:  # 4116
        return {"water_leak": data[0]}  # 1 => on => wet

    elif eid == 0x1015 and length == 1:  # 4117
        # TODO: equipment failure (0x02)
        return {"smoke": data[0]}  # 1 => on => alarm

    elif eid == 0x1016 and length == 1:  # 4118
        return {"gas": data[0]}  # 1 => on => alarm

    elif eid == 0x1017 and length == 4:  # 4119
        # The duration of the unmanned state, in seconds
        return {"idle_time": int.from_bytes(data, "little")}

    elif eid == 0x1018 and length == 1:  # 4120
        # Door Sensor 2: 0 - dark, 1 - light
        return {"light": 1 if data[0] else 0}

    elif eid == 0x1019 and length == 1:  # 4121
        # 0x00: open the door, 0x01: close the door,
        # 0x02: not closed after timeout, 0x03: device reset
        # 1 => true => open
        if data[0] == 0:
            return {"contact": 1}
        elif data[0] == 1:
            return {"contact": 0}
        else:
            return {}

    elif eid == 0x0006 and len(data) == 5:
        action = int.from_bytes(data[4:], "little")
        if action >= len(BLE_FINGERPRINT_ACTION):
            return None
        # status, action, state
        return {
            "action": "fingerprint",
            "action_id": action,
            "key_id": hex(int.from_bytes(data[:4], "little")),
            "message": BLE_FINGERPRINT_ACTION[action],
        }

    elif eid == 0x0007:
        # TODO: lock timestamp
        if data[0] >= len(BLE_DOOR_ACTION):
            return None
        return {
            "action": "door",
            "action_id": data[0],
            "message": BLE_DOOR_ACTION[data[0]],
        }

    elif eid == 0x0008:
        # TODO: lock timestamp
        return {"action": "armed", "state": bool(data[0])}

    elif eid == 0x000B:  # 11
        action = data[0] & 0x0F
        method = data[0] >> 4
        key_id = int.from_bytes(data[1:5], "little")
        error = BLE_LOCK_ERROR.get(key_id)

        # all keys except Bluetooth have only 65536 values
        if error is None and method > 0:
            key_id &= 0xFFFF
        elif error:
            key_id = hex(key_id)

        timestamp = int.from_bytes(data[5:], "little")
        timestamp = datetime.fromtimestamp(timestamp).isoformat()

        if action not in BLE_LOCK_ACTION or method not in BLE_LOCK_METHOD:
            return None

        return {
            "action": "lock",
            "action_id": action,
            "method_id": method,
            "message": BLE_LOCK_ACTION[action],
            "method": BLE_LOCK_METHOD[method],
            "key_id": key_id,
            "error": error,
            "timestamp": timestamp,
        }

    elif eid == 0x0F:  # 15
        # Night Light 2: 1 - moving no light, 100 - moving with light
        # Motion Sensor 2: 0 - moving no light, 256 - moving with light
        # Qingping Motion Sensor - moving with illuminance data
        value = int.from_bytes(data, "little")
        return (
            {"motion": 1, "illuminance": value}
            if pdid == 2691
            else {"motion": 1, "light": int(value >= 100)}
        )

    elif eid == 0x10 and len(data) == 2:  # 16
        # Toothbrush Т500
        if data[0] == 0:
            return {"action": "start", "counter": data[1]}
        else:
            return {"action": "finish", "score": data[1]}

    return None


def get_device(pdid: int, default_name: str) -> Optional[dict]:
    for device in DEVICES:
        if pdid in device:
            desc = device[pdid]
            return {
                "device_manufacturer": desc[0],
                "device_name": desc[0] + " " + desc[1],
                "device_model": desc[2] if len(desc) > 2 else str(pdid),
                "lumi_spec": None,
                "miot_spec": device.get("miot_spec"),
                # if color temp not default 2700..6500
                "color_temp": COLOR_TEMP.get(pdid),
                "max_brightness": MAX_BRIGHTNESS.get(pdid),
            }

    return (
        {
            "device_name": default_name,
            "device_model": str(pdid),
            "lumi_spec": None,
            # default Mesh device will be Bulb
            "miot_spec": [
                [2, 1, "light", "light"],
                [2, 2, "brightness", None],
                [2, 3, "color_temp", None],
            ],
        }
        if default_name == "Mesh"
        else {
            "device_name": default_name,
            "device_model": str(pdid),
        }
    )



import asyncio
import logging

from bleak import discover
from fhempy.lib.generic import FhemModule

from .. import fhem


class discover_ble(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        # disable bleak discovery messages
        logging.getLogger("bleak.backends.bluezdbus.discovery").setLevel(logging.ERROR)
        self.hash = None
        self.blescanTask = None

    async def runBleScan(self):
        while True:
            try:
                devices = await discover()
                for d in devices:
                    if d.name == "GfBT Project":
                        if not await fhem.checkIfDeviceExists(
                            self.hash, "TYPE", "GFPROBT", "MAC", d.address
                        ):
                            self.logger.debug(
                                "create device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                            await fhem.CommandDefine(
                                self.hash,
                                d.name
                                + "_"
                                + d.address.replace(":", "")
                                + " GFPROBT '"
                                + d.address
                                + "'",
                            )
                        else:
                            self.logger.debug(
                                "existing device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                    elif d.name == "CC-RT-BLE":
                        if not await fhem.checkIfDeviceExists(
                            self.hash, "PYTHONTYPE", "eq3bt", "MAC", d.address
                        ):
                            self.logger.debug(
                                "create device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                            await fhem.CommandDefine(
                                self.hash,
                                d.name
                                + "_"
                                + d.address.replace(":", "")
                                + " PythonModule eq3bt '"
                                + d.address
                                + "'",
                            )
                        else:
                            self.logger.debug(
                                "existing device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                    elif d.name[0:7] == "Expert_":
                        if not await fhem.checkIfDeviceExists(
                            self.hash, "PYTHONTYPE", "nespresso_ble", "MAC", d.address
                        ):
                            self.logger.debug(
                                "create device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                            await fhem.CommandDefine(
                                self.hash,
                                d.name
                                + "_"
                                + d.address.replace(":", "")
                                + " PythonModule nespresso_ble '"
                                + d.address
                                + "'",
                            )
                        else:
                            self.logger.debug(
                                "existing device: "
                                + d.name
                                + " / "
                                + d.address
                                + " / rssi: "
                                + str(d.rssi)
                            )
                    else:
                        self.logger.debug(
                            "found unhandled device: "
                            + d.name
                            + ", "
                            + d.address
                            + ", rssi: "
                            + str(d.rssi)
                        )
            except Exception:
                self.logger.error("BLE Scan failed, retry in 600s", exc_info=True)
            await asyncio.sleep(600)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "active")
        await fhem.readingsEndUpdate(hash, 1)

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

        if self.blescanTask:
            self.blescanTask.cancel()

        self.blescanTask = self.create_async_task(self.runBleScan())



import asyncio
import concurrent.futures
import functools
import random
import time
from datetime import datetime
from enum import IntEnum

from dbus import DBusException

from .. import fhem, utils
from ..generic import FhemModule
from . import eq3btsmart as eq3
from .connection import BTLEConnection


class Mode(IntEnum):
    """Thermostat modes."""

    Unknown = -1
    Closed = 0
    Open = 1
    Auto = 2
    Manual = 3
    Away = 4
    Boost = 5


# TODO set schedules
# TODO set windowOpen, windowOpenTime, eco/comfortTemperature


class eq3bt(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_list = {
            "keep_connected": {
                "default": "on",
                "format": "str",
                "options": "on,off",
                "help": "On...keeps bluetooth low energy connection active which makes commands to be executed immediately",
            }
        }
        self.set_attr_config(attr_list)

        set_list_conf = {
            "on": {},
            "off": {},
            "desiredTemperature": {
                "args": ["target_temp"],
                "options": "slider,4.5,0.5,30,1",
            },
            "updateStatus": {},
            "boost": {
                "args": ["target_state"],
                "options": "on,off",
                "help": "Activate boost for 300s",
            },
            "mode": {"args": ["target_mode"], "options": "manual,automatic"},
            "eco": {},
            "comfort": {},
            "childlock": {"args": ["target_state"], "options": "on,off"},
            "resetConsumption": {
                "args": ["cons_var"],
                "options": "all,consumption,consumptionToday,consumptionYesterday",
            },
            "temperatureOffset": {
                "args": ["offset"],
                "params": {"offset": {"format": "float"}},
                "options": "slider,-3.5,0.5,3.5,1",
            },
            "windowOpenTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,4.5,0.5,30,1",
            },
            "windowOpenTime": {
                "args": ["seconds"],
                "params": {"seconds": {"format": "int"}},
                "options": "slider,0,1,3600,1",
            },
            "ecoTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,4.5,0.5,30,1",
            },
            "comfortTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,4.5,0.5,30,1",
            },
        }
        self.set_set_config(set_list_conf)

        self._last_update = 0
        self._mac = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash
        if len(args) < 4:
            return "Usage: define eq3_livingroom fhempy eq3bt <MAC>"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self.logger.info(f"Define: eq3bt {self._mac}")

        icon = await fhem.AttrVal(self.hash["NAME"], "icon", "noicon")
        if icon == "noicon":
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " icon sani_heating_temp"
            )
        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", "offline", 1)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connecting", 1)

        # handle missing dbus configuration
        try:
            self.thermostat = FhemThermostat(
                self.logger,
                self._mac,
                keep_connection=self._attr_keep_connected == "on",
            )
        except DBusException:
            dbus_conf_err = (
                "Please add following configuration to /etc/dbus-1/system.d/bluetooth.conf:\n"
                '<policy user="fhem">\n'
                '    <allow own="org.bluez"/>\n'
                '    <allow send_destination="org.bluez"/>\n'
                '    <allow send_interface="org.bluez.GattCharacteristic1"/>\n'
                '    <allow send_interface="org.bluez.GattDescriptor1"/>\n'
                '    <allow send_interface="org.freedesktop.DBus.ObjectManager"/>\n'
                '    <allow send_interface="org.freedesktop.DBus.Properties"/>\n'
                "</policy>\n\n"
                "ATTENTION: On remote device change the user account above to the one which runs fhempy (e.g. pi)\n\n"
                "Restart dbus afterwards: sudo systemctl restart dbus"
            )
            self.logger.error(dbus_conf_err)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "state", dbus_conf_err, 1
            )
            return dbus_conf_err

        self.create_async_task(self.check_online())
        self.create_async_task(self.consumption_rotate())
        return ""

    def seconds_till_midnight(self):
        """Get the number of seconds until midnight."""
        n = datetime.now()
        return (
            ((24 - n.hour - 1) * 60 * 60) + ((60 - n.minute - 1) * 60) + (60 - n.second)
        )

    async def consumption_rotate(self):
        while True:
            await asyncio.sleep(self.seconds_till_midnight())
            consumption = float(
                await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", consumption, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", "0", 1
            )

    async def set_resetConsumption(self, hash, params):
        cons_var = params["cons_var"]
        if cons_var == "all":
            await fhem.readingsSingleUpdateIfChanged(self.hash, "consumption", 0, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", 0, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", 0, 1
            )
        else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, cons_var, 0, 1)

    async def set_attr_keep_connected(self, hash):
        self.thermostat.set_keep_connected(self._attr_keep_connected == "on")

    async def check_online(self):
        waittime = 300
        if self._attr_keep_connected == "on":
            waittime = 60
        await asyncio.sleep(int(random.random() * 100))
        while True:
            try:
                if time.time() - self._last_update > (waittime * 4):
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "presence", "offline", 1
                    )
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "update", 1
                    )
                await self.update_all()
            except Exception:
                self.logger.error(f"Failed to update, retry in {waittime}s")
            await asyncio.sleep(waittime)

    async def update_all(self):
        self.logger.debug("start update_all")
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await asyncio.get_event_loop().run_in_executor(
                pool, functools.partial(self.thermostat.update_all)
            )
        await self.update_all_readings()

    async def update_all_readings(self):
        await self.update_readings()
        await self.update_id_readings()
        await self.update_schedule_readings()

    async def update_readings(self):
        old_valve_pos = float(
            await fhem.ReadingsVal(self.hash["NAME"], "valvePosition", "0")
        )
        old_consumption = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumption", "0")
        )
        old_consumption_today = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
        )
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "battery", self.thermostat.battery
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "boost", self.thermostat.boost
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "childlock", self.thermostat.locked
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "desiredTemperature", self.thermostat.target_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "ecoTemperature", self.thermostat.eco_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "temperatureOffset", self.thermostat.temperature_offset
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "comfortTemperature", self.thermostat.comfort_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mode", self.thermostat.fhem_mode
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "state", self.thermostat.state
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "completeState", self.thermostat.mode_readable
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "valvePosition", self.thermostat.valve_state
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "awayEnd", self.thermostat.away_end
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpen", self.thermostat.window_open
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpenTemperature", self.thermostat.window_open_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpenTime", self.thermostat.window_open_time
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "presence", "online")
        if (time.time() - self._last_update) < 400:
            consumption_diff = (
                (old_valve_pos + self.thermostat.valve_state)
                / 2
                / 100
                * (time.time() - self._last_update)
                / 60
            )
        else:
            consumption_diff = 0
        new_consumption = round(old_consumption + consumption_diff, 2)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "consumption", new_consumption
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash,
            "consumptionToday",
            round(old_consumption_today + consumption_diff, 2),
        )
        await fhem.readingsEndUpdate(self.hash, 1)
        self._last_update = time.time()

    async def update_id_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "firmware", self.thermostat.firmware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "serialNumber", self.thermostat.device_serial
        )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_schedule_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        for day in self.thermostat.schedule.keys():
            reading = f"schedule_{day}_1"
            if self.thermostat.schedule[day].base_temp == 0 or isinstance(
                self.thermostat.schedule[day].next_change_at, int
            ):
                await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                last_change = "00:00"
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    reading,
                    f"00:00 - {self.thermostat.schedule[day].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].base_temp}",
                )
                last_change = self.thermostat.schedule[day].next_change_at.strftime(
                    "%H:%M"
                )
            last_schedule = False
            for h in range(0, 6):
                reading = f"schedule_{day}_{h+2}"
                if (
                    h == 6
                    or self.thermostat.schedule[day].hours[h].target_temp == 0
                    or isinstance(
                        self.thermostat.schedule[day].hours[h].next_change_at, int
                    )
                    or last_schedule
                ):
                    if last_schedule:
                        await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                    else:
                        value = f"{last_change} - 00:00: {self.thermostat.schedule[day].base_temp}"
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash, reading, value
                        )
                    last_schedule = True
                else:
                    value = f"{last_change} - {self.thermostat.schedule[day].hours[h].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].hours[h].target_temp}"
                    last_change = (
                        self.thermostat.schedule[day]
                        .hours[h]
                        .next_change_at.strftime("%H:%M")
                    )
                    await fhem.readingsBulkUpdateIfChanged(self.hash, reading, value)
        await fhem.readingsEndUpdate(self.hash, 1)

    async def set_and_update(self, fct):
        await utils.run_blocking(fct)
        await self.update_readings()

    def string_to_seconds(self, timestr):
        ftr = [3600, 60, 1]
        return sum([a * b for a, b in zip(ftr, map(int, timestr.split(":")))])

    # SET Functions BEGIN
    async def set_on(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, 30)
            )
        )

    async def set_off(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, 4.5)
            )
        )

    async def set_desiredTemperature(self, hash, params):
        temp = float(params["target_temp"])
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, temp)
            )
        )

    async def set_ecoTemperature(self, hash, params):
        comfort_temp = float(
            await fhem.ReadingsVal(self.hash["NAME"], "comfortTemperature", "17.0")
        )
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_temperature_presets,
                    comfort_temp,
                    params["temp"],
                )
            )
        )

    async def set_comfortTemperature(self, hash, params):
        eco_temp = float(
            await fhem.ReadingsVal(self.hash["NAME"], "ecoTemperature", "15.0")
        )
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_temperature_presets, params["temp"], eco_temp
                )
            )
        )

    async def set_temperatureOffset(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_temperature_offset, params["offset"]
                )
            )
        )

    async def set_windowOpenTemperature(self, hash, params):
        duration = await fhem.ReadingsVal(
            self.hash["NAME"], "windowOpenTime", "0:15:00"
        )
        duration_sec = self.string_to_seconds(duration)
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_windows_open_config,
                    params["temp"],
                    duration_sec,
                )
            )
        )

    async def set_windowOpenTime(self, hash, params):
        temp = await fhem.ReadingsVal(self.hash["NAME"], "windowOpenTemperature", "5")
        temp = float(temp)
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_windows_open_config, temp, params["seconds"]
                )
            )
        )

    async def set_updateStatus(self, hash, params):
        self.create_async_task(self.update_all())

    async def set_boost(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_boost, params["target_state"] == "on"
                )
            )
        )

    async def set_mode(self, hash, params):
        target_mode = params["target_mode"]
        if target_mode == "automatic":
            target_mode = eq3.Mode.Auto
        else:
            target_mode = eq3.Mode.Manual
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_fhem_mode, target_mode)
            )
        )

    async def set_eco(self, hash, params):
        self.create_async_task(
            self.set_and_update(functools.partial(self.thermostat.activate_eco))
        )

    async def set_comfort(self, hash, params):
        self.create_async_task(
            self.set_and_update(functools.partial(self.thermostat.activate_comfort))
        )

    async def set_childlock(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_locked, params["target_state"] == "on"
                )
            )
        )

    # SET Functions END


class FhemThermostat(eq3.Thermostat):
    def __init__(self, logger, mac, keep_connection):
        self.logger = logger
        self._keep_conection = keep_connection
        super(FhemThermostat, self).__init__(mac, BTLEConnection, keep_connection=True)

    def set_keep_connection(self, new_state):
        self.set_keep_connected(new_state)

    def update_all(self):
        super().update()
        super().query_id()
        for day in range(0, 6):
            super().query_schedule(day)

    def set_temperature_presets(self, comfort_temp, eco_temp):
        self.temperature_presets(comfort_temp, eco_temp)

    def set_temperature_offset(self, temp):
        self.temperature_offset = temp

    def set_windows_open_config(self, temperature, duration):
        self.window_open_config(temperature, duration)

    def set_target_temperature(self, temp):
        self.target_temperature = temp

    def set_boost(self, state):
        self.boost = state

    def set_locked(self, state):
        self.locked = state

    def set_fhem_mode(self, mode):
        self.mode = mode

    @property
    def fhem_mode(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "manual"
        elif self._mode == Mode.Open:
            return "manual"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def state(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "off"
        elif self._mode == Mode.Open:
            return "on"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def battery(self):
        if self.low_battery:
            return "low"
        return "ok"



import json
import logging
import re
import socket
import time
from pathlib import Path
from threading import Thread
from typing import Optional

import yaml

from paho.mqtt.client import Client, MQTTMessage
from . import bluetooth, utils, zigbee
from .helpers import DevicesRegistry
from .mini_miio import SyncmiIO
from .shell import TelnetShell, ntp_time
from .unqlite import Unqlite, SQLite

_LOGGER = logging.getLogger(__name__)

RE_NWK_KEY = re.compile(r"lumi send-nwk-key (0x.+?) {(.+?)}")

TELNET_CMD = '{"method":"enable_telnet_service","params":""}'

GW3_ERROR = (
    "Check if [issue](https://github.com/AlexxIT/XiaomiGateway3/issues) "
    "exists or create new one:\n```\n%s\n```"
)


class GatewayBase(DevicesRegistry):
    did: str = None
    """Xiaomi did of the gateway"""

    available: bool = None
    """Getaway is considered online if there is an active connection to mqtt"""
    enabled: bool = None
    """Gateway stops main loop if enabled property sets to False"""

    host: str = None

    options: dict = None

    # TODO: remove this prop
    gw_topic: str = None

    mqtt: Client = None
    miio: SyncmiIO = None

    @property
    def debug_mode(self):
        return self.options.get("debug", "")

    @property
    def device(self):
        return self.devices[self.did]

    def debug(self, message: str):
        # basic logs
        if "true" in self.debug_mode:
            _LOGGER.debug(f"{self.host} | {message}")


class GatewayMesh(GatewayBase):
    mesh_params: list = None
    mesh_ts: float = 0

    def mesh_start(self):
        self.mesh_params = []

        for device in self.devices.values():
            if device["type"] == "mesh" and "childs" not in device:
                # TODO: rewrite more clear logic for lights and switches
                p = device["miot_spec"][0]
                self.mesh_params.append(
                    {"did": device["did"], "siid": p[0], "piid": p[1]}
                )

        if self.mesh_params:
            self.mesh_ts = time.time() + 30
            Thread(target=self.mesh_run, name=f"{self.host}_mesh").start()

    def mesh_run(self):
        self.debug("Start Mesh Thread")

        while self.enabled:
            if time.time() < self.mesh_ts:
                time.sleep(1)
                continue

            try:
                resp = self.miio.send_bulk("get_properties", self.mesh_params)
                if resp:
                    params2 = []
                    for item in resp:
                        if "value" not in item:
                            continue

                        did = item["did"]
                        device_params = self.devices[did]["miot_spec"]
                        # get other props for turn on lights or live switches
                        if device_params[0][3] == "switch" or item["value"]:
                            params2 += [
                                {"did": did, "siid": p[0], "piid": p[1]}
                                for p in device_params[1:]
                            ]

                    if params2:
                        resp2 = self.miio.send_bulk("get_properties", params2)
                        if resp2:
                            resp += resp2

                    self.debug(f"Pull Mesh {resp}")
                    self.process_mesh_data(resp)

                else:
                    self.debug("Can't get mesh bulb state")

            except Exception as e:
                self.debug(f"ERROR in mesh thread {e}")

            self.mesh_ts = time.time() + 30

    def process_mesh_data(self, data: list):
        """Can receive multiple properties from multiple devices.

        data = [{'did':123,'siid':2,'piid':1,'value:True}]
        """
        bulk = {}

        for param in data:
            code = param.get("code", 0)
            if code == -4004:
                # handle device offline state
                param["value"] = None
            elif code != 0:
                continue

            did = param["did"]
            if did not in self.devices:
                continue

            device = self.devices[did]

            prop = next(
                (
                    p[2]
                    for p in device["miot_spec"]
                    if p[0] == param["siid"] and p[1] == param["piid"]
                ),
                None,
            )
            if not prop:
                continue

            if did not in bulk:
                bulk[did] = {}

            # TODO: fix this dirty hack
            if device["model"] == 3083 and prop == "power" and param["value"]:
                param["value"] /= 100.0

            # https://github.com/AlexxIT/XiaomiGateway3/issues/312
            if prop in ("temperature", "humidity") and param["value"]:
                param["value"] = round(param["value"], 2)

            bulk[did][prop] = param["value"]

        for did, payload in bulk.items():
            # self.debug(f"Process Mesh Data for {did}: {payload}")
            device = self.devices[did]
            for entity in device["entities"].values():
                if entity:
                    entity.update(payload)

    def send_mesh(self, device: dict, data: dict):
        # data = {'light':True}
        payload = []
        for k, v in data.items():
            param = next(p for p in device["miot_spec"] if p[2] == k)
            payload.append(
                {
                    "did": device["did"],
                    "siid": param[0],
                    "piid": param[1],
                    "value": v if param[0] != 8 else int(v),
                }
            )

        try:
            # 2 seconds are selected experimentally
            if self.miio.send("set_properties", payload):
                self.mesh_force_update()
        except Exception:
            self.debug(f"Can't send mesh {device['did']} => {data}")

    def mesh_force_update(self):
        self.mesh_ts = time.time() + 2


class GatewayStats(GatewayMesh):
    # global stats for all gateways
    # stats: dict = {}
    parent_scan_ts: float = 0

    # interval for auto parent refresh in minutes, 0 - disabled auto refresh
    # -1 - disabled
    parent_scan_interval: Optional[int] = None

    # collected data from MQTT topic log/z3 (zigbee console)
    z3buffer: Optional[dict] = None

    @property
    def stats_enable(self):
        return self.options.get("stats")

    def process_gw_stats(self, payload: dict = None):
        # empty payload - update available state
        self.debug(f"gateway <= {payload or self.available}")

        if self.parent_scan_ts and time.time() > self.parent_scan_ts:
            # block any auto updates in 30 seconds
            self.parent_scan_ts = time.time() + 30

            self.get_gateway_info()

        device = self.devices.get(self.did)
        if not device or not device.get("stats"):
            return

        if payload:
            if "networkUp" in payload:
                # {"networkUp":false}
                payload = {
                    "network_pan_id": payload.get("networkPanId"),
                    "radio_tx_power": payload.get("radioTxPower"),
                    "radio_channel": payload.get("radioChannel"),
                }
            elif "free_mem" in payload:
                s = payload["run_time"]
                d, h, m, s = (
                    s // (3600 * 24),
                    s % (3600 * 24) // 3600,
                    s % 3600 // 60,
                    s % 60,
                )

                payload = {
                    "free_mem": payload["free_mem"],
                    "load_avg": payload["load_avg"],
                    "rssi": -payload["rssi"],
                    "uptime": f"{d} days, {h:02}:{m:02}:{s:02}",
                }

        device["stats"].update(payload)

    def process_zb_stats(self, payload: dict) -> bool:
        # convert ieee to did
        did = "lumi." + str(payload["eui64"]).lstrip("0x").lower()
        device = self.devices.get(did)
        if device and device.get("stats"):
            device["stats"].update(payload)
            return True
        return False

    def process_ble_stats(self, mac: str, data: dict = None):
        device = self.devices.get(mac)
        if device and device.get("stats"):
            device["stats"].update(data)

    def process_z3(self, payload: str):
        if payload.startswith("CLI command executed"):
            cmd = payload[22:-1]
            if cmd == "debugprint all_on" or self.z3buffer is None:
                # reset all buffers
                self.z3buffer = {}
            else:
                self.z3buffer[cmd] = self.z3buffer["buffer"]

            self.z3buffer["buffer"] = ""

            if cmd == "plugin concentrator print-table":
                self._process_gateway_info()

        elif self.z3buffer:
            self.z3buffer["buffer"] += payload

    def get_gateway_info(self):
        self.debug("Update zigbee network info")
        payload = {
            "commands": [
                {"commandcli": "debugprint all_on"},
                {"commandcli": "plugin device-table print"},
                {"commandcli": "plugin stack-diagnostics child-table"},
                {"commandcli": "plugin stack-diagnostics neighbor-table"},
                {"commandcli": "plugin concentrator print-table"},
                {"commandcli": "debugprint all_off"},
            ]
        }
        payload = json.dumps(payload, separators=(",", ":"))
        self.mqtt.publish(self.gw_topic + "commands", payload)

    def send_zigbee_cli(self, commands: list):
        payload = {"commands": [{"commandcli": cmd} for cmd in commands]}
        payload = json.dumps(payload, separators=(",", ":"))
        self.mqtt.publish(self.gw_topic + "commands", payload)

    def _process_gateway_info(self):
        self.debug("Update parent info table")

        try:
            raw = self.z3buffer["plugin device-table print"]
            m1 = re.findall(
                r"\d+ ([A-F0-9]{4}): {2}([A-F0-9]{16}) 0 {2}\w+ " r"(\d+)", raw
            )

            raw = self.z3buffer["plugin stack-diagnostics child-table"]
            m2 = re.findall(r"\(>\)([A-F0-9]{16})", raw)

            raw = self.z3buffer["plugin stack-diagnostics neighbor-table"]
            m3 = re.findall(r"\(>\)([A-F0-9]{16})", raw)

            raw = self.z3buffer["plugin concentrator print-table"]
            m4 = re.findall(r": ([A-F0-9x> -]{16,}) -> 0x0000", raw)
            m4 = [i.replace("0x", "").split(" -> ") for i in m4]
            m4 = {i[0]: i[1:] for i in m4}

            self.debug(f"Total zigbee devices: {len(m1)}")

            for i in m1:
                ieee = "0x" + i[1]

                nwk = i[0]
                ago = int(i[2])
                type_ = "device" if i[1] in m2 else "router" if i[1] in m3 else "-"
                parent = "0x" + m4[nwk][0] if nwk in m4 else "-"

                payload = {
                    "eui64": ieee,
                    "nwk": "0x" + nwk,
                    "ago": ago,
                    "type": type_,
                    "parent": parent,
                }

                if not self.process_zb_stats(payload):
                    self.debug(f"Unknown zigbee device {ieee}: {payload}")

            # one hour later
            if self.parent_scan_interval > 0:
                self.parent_scan_ts = time.time() + self.parent_scan_interval * 60

        except Exception as e:
            self.debug(f"Can't update parents: {e}")


class GatewayGW3(GatewayStats):
    ble_known = {}
    ble_retain = {}

    def process_gw3_state(self, mac: str, data: dict):
        mac = mac.lower().replace(":", "")
        if mac not in self.devices:
            # save data for first init with device/info topic
            self.ble_retain[mac] = data
            return

        device = self.devices[mac]

        # skip messages with same sequence numb
        if "seq" in data:
            if device.get("seq") == data["seq"]:
                return
            device["seq"] = data["seq"]

        # don't create stats if device don't have entities
        if self.stats_enable and device["entities"]:
            self.add_stats(device)
            self.process_ble_stats(mac)

        self.process_ble_payload(device, data)

    def process_gw3_info(self, mac: str, data: dict):
        if data["type"] != "ble":
            return

        mac = mac.lower().replace(":", "")
        # defaults can be set from: YAML, Cloud and integrations GUI
        if mac not in self.defaults:
            name = f"{data['brand']} {data['name']} ({mac.upper()})"
            self.debug(f"Skip new BLE device: {name}")
            self.ble_known[mac] = name
            return

        device = self.find_or_create_device(
            {
                "mac": mac,
                "type": "ble",
                "model": None,
                "init": {},
                "device_manufacturer": data["brand"],
                "device_name": data["brand"] + " " + data["name"],
                "device_model": data["model"],
            }
        )

        if mac in self.ble_retain:
            self.process_ble_payload(device, self.ble_retain[mac])

    def process_ble_payload(self, device: dict, payload: dict):
        # init entities if needed
        for k, v in payload.items():
            if k in device["entities"]:
                continue

            device["init"][k] = payload[k]

            domain = bluetooth.get_ble_domain(k, v)
            self.add_entity(domain, device, k)

        for entity in device["entities"].values():
            if entity:
                entity.update(payload)


# noinspection PyUnusedLocal
class GatewayEntry(Thread, GatewayGW3):
    """Main class for working with the gateway via Telnet (23), MQTT (1883) and
    miIO (54321) protocols.
    """

    time_offset = 0
    pair_model = None
    pair_payload = None
    pair_payload2 = None

    def __init__(self, host: str, token: str, **options):
        super().__init__(daemon=True, name=f"{host}_main")

        self.host = host
        self.options = options

        self.miio = SyncmiIO(host, token)

        self.mqtt = Client()
        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect
        self.mqtt.on_message = self.on_message

        if "true" in self.debug_mode:
            self.miio.debug = True

        self.setups = {}

    @property
    def ble_mode(self):
        return self.options.get("ble", True)

    @property
    def parent_scan_interval(self):
        return self.options.get("parent", -1)

    @property
    def telnet_cmd(self):
        return self.options.get("telnet_cmd") or TELNET_CMD

    @property
    def hass(self):
        for device in self.devices.values():
            for entity in device["entities"].values():
                if entity and entity.hass:
                    return entity.hass
        return None

    def stop(self, *args):
        self.enabled = False
        self.mqtt.disconnect()

        for device in self.devices.values():
            if self in device["gateways"]:
                device["gateways"].remove(self)

    def run(self):
        """Main thread loop."""
        self.debug("Start main thread")

        self.mqtt.connect_async(self.host)

        self.enabled = True
        while self.enabled:
            # if not telnet - enable it
            if not self._check_port(23) and not self._enable_telnet():
                time.sleep(30)
                continue

            if not self.did:
                devices = self._get_devices()
                if not devices:
                    time.sleep(60)
                    continue

                self.setup_devices(devices)
                self.update_time_offset()
                self.mesh_start()

            # if not mqtt - enable it (handle Mi Home and ZHA mode)
            if not self._prepare_gateway() or not self._mqtt_connect():
                time.sleep(60)
                continue

            self.mqtt.loop_forever()

        self.debug("Stop main thread")

    def update_time_offset(self):
        gw_time = ntp_time(self.host)
        if gw_time:
            self.time_offset = gw_time - time.time()
            self.debug(f"Gateway time offset: {self.time_offset}")

    def _check_port(self, port: int):
        """Check if gateway port open."""
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            return s.connect_ex((self.host, port)) == 0
        finally:
            s.close()

    def _enable_telnet(self):
        """Enable telnet with miio protocol."""
        raw = json.loads(self.telnet_cmd)
        if self.miio.send(raw["method"], raw.get("params")) != ["ok"]:
            self.debug("Can't enable telnet")
            return False
        return True

    def _prepare_gateway(self):
        """Launching the required utilities on the hub, if they are not already
        running.
        """
        self.debug("Prepare Gateway")
        try:
            shell = TelnetShell(self.host)
            self.debug(f"Version: {shell.ver}")

            ps = shell.get_running_ps()

            if "mosquitto -d" not in ps:
                self.debug("Run public mosquitto")
                shell.run_public_mosquitto()

            if "ntpd" not in ps:
                # run NTPd for sync time
                shell.run_ntpd()

            if "-t log/miio" not in ps:
                # all data or only necessary events
                pattern = (
                    '\\{"'
                    if "miio" in self.debug_mode
                    else "ot_agent_recv_handler_one.+" "properties_changed|heartbeat"
                )
                self.debug("Redirect miio to MQTT")
                shell.redirect_miio2mqtt(pattern)

            if self.options.get("buzzer"):
                if "dummy:basic_gw" not in ps:
                    self.debug("Disable buzzer")
                    shell.stop_buzzer()
            else:
                if "dummy:basic_gw" in ps:
                    self.debug("Enable buzzer")
                    shell.run_buzzer()

            # running gw3 last because it touches the daemon_miio.sh file
            if self.ble_mode:
                # check if binary has latest version
                if shell.check_gw3():
                    if "/gw3" not in ps:
                        self.debug("Run gw3")
                        shell.run_gw3()
                else:
                    if "/gw3" in ps:
                        shell.stop_gw3()
                    self.debug("Download and run gw3")
                    shell.run_gw3()
            elif "/gw3" in ps:
                self.debug("Stop gw3")
                shell.stop_gw3()

            if self.options.get("zha"):
                # stop lumi without checking if it's running
                shell.stop_lumi_zigbee()

                if "tcp-l:8889" in ps:
                    shell.stop_zigbee_tcp()
                if "tcp-l:8888" not in ps:
                    self.debug("Run Zigbee TCP")
                    shell.run_zigbee_tcp()

            else:
                # check both 8888 and 8889
                if "tcp-l:888" in ps:
                    self.debug("Stop Zigbee TCP")
                    shell.stop_zigbee_tcp()

                if self.parent_scan_interval >= 0:
                    if "Lumi_Z3GatewayHost_MQTT -n 1 -b 115200 -l" not in ps:
                        self.debug("Run public Zigbee console")
                        shell.run_public_zb_console()

                    if self.parent_scan_interval > 0:
                        self.parent_scan_ts = 1

                elif "daemon_app.sh" not in ps:
                    self.debug("Run Lumi Zigbee")
                    shell.run_lumi_zigbee()

            shell.close()

            return True

        except (ConnectionRefusedError, socket.timeout):
            return False

        except Exception as e:
            self.debug(f"Can't prepare gateway: {e}")
            return False

    def _mqtt_connect(self) -> bool:
        try:
            self.mqtt.reconnect()
            return True
        except Exception:
            return False

    def _get_devices(self):
        """Load devices info for Coordinator, Zigbee and Mesh."""
        try:
            shell = TelnetShell(self.host)

            # 1. Read coordinator info
            raw = shell.read_file("/data/zigbee/coordinator.info")
            device = json.loads(raw)
            devices = [
                {
                    "did": shell.get_did(),
                    "model": "lumi.gateway.mgl03",
                    "mac": device["mac"],
                    "wlan_mac": shell.get_wlan_mac(),
                    "type": "gateway",
                    "fw_ver": shell.ver,
                    "online": True,
                    "init": {
                        "firmware lock": shell.check_firmware_lock(),
                    },
                }
            ]

            # 2. Read zigbee devices
            if not self.options.get("zha"):
                # read Silicon devices DB
                nwks = {}
                try:
                    raw = shell.read_file("/data/silicon_zigbee_host/devices.txt")
                    raw = raw.decode().split(" ")
                    for i in range(0, len(raw) - 1, 32):
                        ieee = reversed(raw[i + 3 : i + 11])
                        ieee = "".join(f"{i:>02s}" for i in ieee)
                        nwks[ieee] = f"{raw[i]:>04s}"
                except Exception:
                    _LOGGER.exception("Can't read Silicon devices DB")

                # read Xiaomi devices DB
                raw = shell.read_file(shell.zigbee_db, as_base64=True)
                # self.debug(f"Devices RAW: {raw}")
                if raw is None:
                    self.debug("No zigbee database")
                    data = {}
                elif raw.startswith(b"unqlite"):
                    db = Unqlite(raw)
                    data = db.read_all()
                else:
                    raw = re.sub(br"}\s*{", b",", raw)
                    data = json.loads(raw)

                # data = {} or data = {'dev_list': 'null'}
                dev_list = json.loads(data.get("dev_list", "null")) or []

                for did in dev_list:
                    model = data.get(did + ".model")
                    if not model:
                        self.debug(f"{did} has not in devices DB")
                        continue
                    desc = zigbee.get_device(model)

                    # skip unknown model
                    if desc is None:
                        self.debug(f"{did} has an unsupported modell: {model}")
                        continue

                    retain = json.loads(data[did + ".prop"])["props"]
                    self.debug(f"{did} {model} retain: {retain}")

                    params = {
                        p[2]: retain.get(p[1])
                        for p in (desc["lumi_spec"] or desc["miot_spec"])
                        if p[1] is not None
                    }

                    ieee = f"{data[did + '.mac']:>016s}"
                    device = {
                        "did": did,
                        "mac": "0x" + data[did + ".mac"],
                        "ieee": ieee,
                        "nwk": nwks.get(ieee),
                        "model": model,
                        "type": "zigbee",
                        "fw_ver": retain.get("fw_ver") or retain.get("1.4"),
                        "init": zigbee.fix_xiaomi_props(model, params),
                        "online": retain.get("alive", 1) == 1,
                    }
                    devices.append(device)

            # 3. Read bluetooth devices
            if self.ble_mode:
                raw = shell.read_file("/data/miio/mible_local.db", as_base64=True)
                try:
                    db = SQLite(raw)

                    # load Mesh groups
                    mesh_groups = {}

                    rows = db.read_table(shell.mesh_group_table)
                    for row in rows:
                        # don't know if 8 bytes enougth
                        mac = int(row[0]).to_bytes(8, "big").hex()
                        device = {
                            "did": "group." + row[0],
                            "mac": mac,
                            "model": 0,
                            "childs": [],
                            "type": "mesh",
                            "online": True,
                        }
                        devices.append(device)

                        group_addr = row[1]
                        mesh_groups[group_addr] = device

                    # load Mesh bulbs
                    rows = db.read_table(shell.mesh_device_table)
                    for row in rows:
                        device = {
                            "did": row[0],
                            "mac": row[1].replace(":", ""),
                            "model": row[2],
                            "type": "mesh",
                            "online": False,
                        }
                        devices.append(device)

                        group_addr = row[5]
                        if group_addr in mesh_groups:
                            # add bulb to group if exist
                            mesh_groups[group_addr]["childs"].append(row[0])

                except Exception:
                    _LOGGER.exception("Can't read mesh devices")

            # for testing purposes
            for k, v in self.defaults.items():
                if k[0] == "_":
                    devices.append(v)

            return devices

        except (ConnectionRefusedError, socket.timeout):
            return None

        except Exception as e:
            _LOGGER.exception(f"{self.host} | Can't read devices: {e}")
            return None

    def lock_firmware(self, enable: bool):
        self.debug(f"Set firmware lock to {enable}")
        try:
            shell = TelnetShell(self.host)
            shell.lock_firmware(enable)
            locked = shell.check_firmware_lock()
            shell.close()
            return enable == locked

        except Exception as e:
            self.debug(f"Can't set firmware lock: {e}")
            return False

    def update_entities_states(self):
        for device in list(self.devices.values()):
            if self in device["gateways"]:
                for entity in list(device["entities"].values()):
                    if entity:
                        entity.schedule_update_ha_state()

    def on_connect(self, client, userdata, flags, rc):
        self.debug("MQTT connected")
        self.mqtt.subscribe("#")

        self.available = True
        self.process_gw_stats()
        # self.update_entities_states()

    def on_disconnect(self, client, userdata, rc):
        self.debug("MQTT disconnected")
        # force end mqtt.loop_forever()
        self.mqtt.disconnect()

        self.available = False
        self.process_gw_stats()
        # self.update_entities_states()

    def on_message(self, client: Client, userdata, msg: MQTTMessage):
        # for debug purpose
        enabled = self.enabled
        try:
            topic = msg.topic

            if "mqtt" in self.debug_mode:
                _LOGGER.debug(f"{self.host} | MQTT | {topic} {msg.payload}")

            if topic == "zigbee/send":
                payload = json.loads(msg.payload)
                self.process_zigbee_message(payload)

            elif topic == "gw3/stderr":
                _LOGGER.warning(msg.payload)

                message = GW3_ERROR % msg.payload.decode("utf-8")
                utils.notification(self.hass, message)

                # restore gw3 application
                self._prepare_gateway()

            elif topic.startswith("gw3/"):
                items = topic.split("/")
                if len(items) != 3:
                    return

                payload = json.loads(msg.payload)
                if items[2] in ("state", "event"):
                    self.process_gw3_state(items[1], payload)
                elif items[2] == "info":
                    self.process_gw3_info(items[1], payload)

            elif topic == "log/miio":
                # don't need to process another data
                if b"ot_agent_recv_handler_one" not in msg.payload:
                    return

                for raw in utils.extract_jsons(msg.payload):
                    if self.ble_mode and b"properties_changed" in raw:
                        data = json.loads(raw)["params"]
                        self.debug(f"Process props {data}")
                        self.process_mesh_data(data)
                    elif b"event.gw.heartbeat" in raw:
                        payload = json.loads(raw)["params"][0]
                        self.process_gw_stats(payload)
                        # time offset may changed right after gw.heartbeat
                        self.update_time_offset()

            elif topic == "log/z3":
                self.process_z3(msg.payload.decode())

            elif topic.endswith("/heartbeat"):
                payload = json.loads(msg.payload)
                self.process_gw_stats(payload)

            elif topic.endswith(("/MessageReceived", "/devicestatechange")):
                payload = json.loads(msg.payload)
                self.process_zb_stats(payload)

            elif self.pair_model and topic.endswith("/commands"):
                self.process_pair(msg.payload)

        except Exception:
            _LOGGER.exception(f"Processing MQTT: {msg.topic} {msg.payload}")

    def setup_devices(self, devices: list):
        """Add devices to hass."""
        for device in devices:
            type_ = device["type"]
            self.debug(f"Setup {type_} device {device}")

            device = self.find_or_create_device(device)

            if type_ == "gateway":
                self.did = device["did"]
                self.gw_topic = f"gw/{device['mac'][2:].upper()}/"

            if type_ in ("gateway", "zigbee", "mesh"):
                for param in device["lumi_spec"] or []:
                    self.add_entity(param[3], device, param[2])
                for param in device["miot_spec"] or []:
                    self.add_entity(param[3], device, param[2])

            if self.stats_enable and type_ in ("gateway", "zigbee"):
                self.add_stats(device)

    def process_zigbee_message(self, data: dict):
        if data["cmd"] == "heartbeat":
            # don't know if only one item
            assert len(data["params"]) == 1, data

            data = data["params"][0]
            pkey = "res_list"
        elif data["cmd"] == "report":
            pkey = "params" if "params" in data else "mi_spec"
        elif data["cmd"] in ("write_rsp", "read_rsp"):
            pkey = "results" if "results" in data else "mi_spec"
        elif data["cmd"] == "write_ack":
            return
        else:
            _LOGGER.warning(f"Unsupported cmd: {data}")
            return

        did = data["did"] if data["did"] != "lumi.0" else self.did

        # skip without callback and without data
        if did not in self.devices or pkey not in data:
            return

        ts = time.time()

        device = self.devices[did]
        payload = {}

        # convert codes to names
        for param in data[pkey]:
            if param.get("error_code", 0) != 0:
                continue

            if "res_name" in param:
                prop = param["res_name"]
            elif "piid" in param:
                prop = f"{param['siid']}.{param['piid']}"
            elif "eiid" in param:
                prop = f"{param['siid']}.{param['eiid']}"
            else:
                _LOGGER.warning(f"Unsupported param: {data}")
                return

            if prop in zigbee.GLOBAL_PROP:
                prop = zigbee.GLOBAL_PROP[prop]
            else:
                prop = next(
                    (
                        p[2]
                        for p in (device["lumi_spec"] or device["miot_spec"])
                        if p[0] == prop
                    ),
                    prop,
                )

            # https://github.com/Koenkk/zigbee2mqtt/issues/798
            # https://www.maero.dk/aqara-temperature-humidity-pressure-sensor-teardown/
            if prop == "temperature" and device["model"] != "lumi.airmonitor.acn01":
                if -4000 < param["value"] < 12500:
                    payload[prop] = param["value"] / 100.0
            elif prop == "humidity" and device["model"] != "lumi.airmonitor.acn01":
                if 0 <= param["value"] <= 10000:
                    payload[prop] = param["value"] / 100.0
            elif prop == "pressure":
                payload[prop] = param["value"] / 100.0
            elif prop == "battery":
                # I do not know if the formula is correct, so battery is more
                # important than voltage
                payload[prop] = zigbee.fix_xiaomi_battery(param["value"])
            elif prop == "alive" and param["value"]["status"] == "offline":
                device["online"] = False
            elif prop == "angle":
                # xiaomi cube 100 points = 360 degrees
                payload[prop] = param["value"] * 4
            elif prop == "duration":
                # xiaomi cube
                payload[prop] = param["value"] / 1000.0
            elif prop in ("consumption", "power"):
                payload[prop] = round(param["value"], 2)
            elif prop in ("voltage", "current"):
                payload[prop] = round(param["value"] / 1000.0, 2)
            elif prop == "energy":
                # energy consumption Wh to kWh
                payload[prop] = round(param["value"] / 1000.0, 3)
            elif prop == "fw_ver" and param["value"] != device["fw_ver"]:
                device["fw_ver"] = param["value"]
                self._update_device_fw_ver(device, param["value"])
            elif prop == "ota_progress":
                self._update_device_fw_ver(device, f"Update {param['value']}%")
            elif "value" in param:
                payload[prop] = param["value"]
            elif "arguments" in param:
                d = yaml.safe_load(prop)
                if isinstance(d, dict):
                    payload.update(d)
                else:
                    payload[prop] = param["arguments"]

        # no time in device add command
        ts = (
            round(ts - data["time"] * 0.001 + self.time_offset, 2)
            if "time" in data
            else "?"
        )
        self.debug(f"{device['did']} {device['model']} <= {payload} [{ts}]")

        if payload:
            device["online"] = True

        for entity in device["entities"].values():
            if entity:
                entity.update(payload)
                break

        # TODO: move code earlier!!!
        if "added_device" in payload:
            # {'did': 'lumi.fff', 'mac': 'fff', 'model': 'lumi.sen_ill.mgl01',
            # 'version': '21', 'zb_ver': '3.0'}
            device = payload["added_device"]
            device["mac"] = "0x" + device["mac"]
            device["type"] = "zigbee"
            device["init"] = payload
            self.setup_devices([device])

        # return for tests purpose
        return payload

    @staticmethod
    def _update_device_fw_ver(device: dict, fw_ver: str):
        for entity in device["entities"].values():
            if entity:
                utils.update_device_info(entity.hass, device["did"], sw_version=fw_ver)
                break

    def process_pair(self, raw: bytes):
        _LOGGER.debug(f"!!! {raw}")
        # get shortID and eui64 of paired device
        if b"lumi send-nwk-key" in raw:
            # create model response
            payload = (
                f"0x08020105000042{len(self.pair_model):02x}"
                f"{self.pair_model.encode().hex()}"
            )
            m = RE_NWK_KEY.search(raw.decode())
            self.pair_payload = json.dumps(
                {
                    "sourceAddress": m[1],
                    "eui64": "0x" + m[2],
                    "profileId": "0x0104",
                    "clusterId": "0x0000",
                    "sourceEndpoint": "0x01",
                    "destinationEndpoint": "0x01",
                    "APSCounter": "0x01",
                    "APSPlayload": payload,
                },
                separators=(",", ":"),
            )
            self.pair_payload2 = json.dumps(
                {
                    "sourceAddress": m[1],
                    "eui64": "0x" + m[2],
                    "profileId": "0x0104",
                    "clusterId": "0x0000",
                    "sourceEndpoint": "0x01",
                    "destinationEndpoint": "0x01",
                    "APSCounter": "0x01",
                    "APSPlayload": "0x0801010100002001",
                },
                separators=(",", ":"),
            )

        # send model response "from device"
        elif b"zdo active " in raw:
            self.mqtt.publish(self.gw_topic + "MessageReceived", self.pair_payload2)
            self.mqtt.publish(self.gw_topic + "MessageReceived", self.pair_payload)

    def send(self, device: dict, data: dict):
        did = device["did"] if device["did"] != self.did else "lumi.0"
        payload = {"cmd": "write", "did": did}

        # convert hass prop to lumi prop
        if device["miot_spec"]:
            try:
                params = []
                for k, v in data.items():
                    k = next(p[0] for p in device["miot_spec"] if p[2] == k)
                    if k == "switch":
                        v = bool(v)
                    params.append({"siid": int(k[0]), "piid": int(k[2]), "value": v})

                # Attention! mi_spec are used in payload
                payload["mi_spec"] = params
            except StopIteration:
                pass

        # lumi_spec and miot_spec persist simultaneously only in gateway
        if device["lumi_spec"]:
            try:
                params = [
                    {
                        "res_name": next(
                            p[0] for p in device["lumi_spec"] if p[2] == k
                        ),
                        "value": v,
                    }
                    for k, v in data.items()
                ]

                payload["params"] = params
            except StopIteration:
                pass

        self.debug(f"{device['did']} {device['model']} => {payload}")
        payload = json.dumps(payload, separators=(",", ":")).encode()
        self.mqtt.publish("zigbee/recv", payload)

    def send_telnet(self, *args: str):
        try:
            shell = TelnetShell(self.host)
            for command in args:
                if command == "ftp":
                    shell.run_ftp()
                elif command == "dump":
                    raw = shell.tar_data()
                    filename = Path().absolute() / f"{self.host}.tar.gz"
                    with open(filename, "wb") as f:
                        f.write(raw)
                else:
                    shell.exec(command)
            shell.close()

        except Exception as e:
            _LOGGER.exception(f"Telnet command error: {e}")

    def send_mqtt(self, cmd: str):
        if cmd == "publishstate":
            self.mqtt.publish(self.gw_topic + "publishstate")

    def get_device(self, mac: str) -> Optional[dict]:
        for device in self.devices.values():
            if device.get("mac") == mac:
                return device
        return None


Gateway3 = GatewayEntry



import asyncio
import time

from fhempy.lib import fhem
from fhempy.lib.generic import FhemModule
from fhempy.lib.xiaomi_gateway3.xiaomi_gateway3 import FhempyGateway

attr_settings = {
    "lumi.sensor_magnet.v2": {
        "devStateIcon": "1:fts_door_open\@red 0:fts_door\@green",
        "stateFormat": "contact",
        "icon": "tuer_fenster_kontakt",
    },
    "lumi.sensor_magnet.aq2": {
        "devStateIcon": "1:fts_door_open\@red 0:fts_door\@green",
        "stateFormat": "contact",
        "icon": "tuer_fenster_kontakt",
    },
    "lumi.sensor_wleak.aq1": {
        "icon": "sani_water_tap",
        "devStateIcon": "0:message_ok 1:humidity@red",
        "stateFormat": "moisture",
    },
    "lumi.sensor_ht.v1": {
        "stateFormat": "temperature °C, humidity %",
        "icon": "temp_temperature",
    },
    "lumi.sensor_ht.v2": {
        "stateFormat": "temperature °C, humidity %, pressure hPa",
        "icon": "temp_temperature",
    },
    "lumi.weather.v1": {
        "stateFormat": "temperature °C, humidity %, pressure hPa",
        "icon": "temp_temperature",
    },
    "lumi.sensor_motion.v1": {
        "stateFormat": "motion",
        "devStateIcon": "1:motion_detector\@red 0:motion_detector\@green",
        "icon": "people_sensor",
    },
    "lumi.sensor_motion.v2": {
        "stateFormat": "motion",
        "devStateIcon": "1:motion_detector\@red 0:motion_detector\@green",
        "icon": "people_sensor",
    },
    "lumi.gateway.mgl03": {
        "devStateIcon": "online:it_wifi\@green offline:it_wifi\@red",
        "icon": "tradfri_gateway",
    },
}


class BaseDevice(FhemModule):
    def __init__(self, logger, gateway: FhempyGateway):
        super().__init__(logger)
        self._gateway = gateway
        self._xg3_device = None
        self.last_update = 0
        self.logger = logger
        self.create_async_task(self.offline_check())

    def set_hash(self, hash):
        self.hash = hash

    @property
    def device(self):
        return self._xg3_device

    async def initialize(self, device):
        self._xg3_device = device
        for attr in attr_settings[device["model"]]:
            if await fhem.AttrVal(self.hash["NAME"], attr, "") == "":
                await fhem.CommandAttr(
                    self.hash,
                    (
                        f"{self.hash['NAME']} {attr} "
                        f"{attr_settings[self._xg3_device['model']][attr]}"
                    ),
                )

        for reading in device:
            # following attributes are not readings
            if reading in ("lumi_spec", "miot_spec", "entities", "gateways", "stats"):
                continue

            if reading == "init":
                await self.async_update(device["init"])
                continue
            elif reading == "params":
                continue
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, reading, device[reading], 1
                )

    def update(self, data):
        asyncio.run_coroutine_threadsafe(self.async_update(data), self.loop).result()

    async def async_update(self, data):
        self.logger.debug(f"update call {str(data)}")
        self.last_update = time.time()
        if data is None:
            return

        await fhem.readingsBeginUpdate(self.hash)
        # device is online
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        # update data
        for reading in data:
            if reading == "added_device":
                pass
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading.replace(" ", "_"), str(data[reading])
                )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def offline_check(self):
        while True:
            if time.time() - self.last_update > 3700:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "offline", 1
                )
            await asyncio.sleep(300)



import base64
import logging
import re
import time
from socket import socket, AF_INET, SOCK_DGRAM
from telnetlib import Telnet
from typing import Union

_LOGGER = logging.getLogger(__name__)

# We should use HTTP-link because wget doesn't support HTTPS and curl removed
# in lastest fw. But it's not a problem because we check md5

# original link http://pkg.musl.cc/socat/mipsel-linux-musln32/bin/socat
# original link https://busybox.net/downloads/binaries/1.21.1/busybox-mipsel
WGET = (
    "(wget http://master.dl.sourceforge.net/project/mgl03/{0}?viasf=1 "
    "-O /data/{1} && chmod +x /data/{1})"
)

RUN_ZIGBEE_TCP = (
    "/data/socat tcp-l:%d,reuseaddr,fork,keepalive,nodelay,"
    "keepidle=1,keepintvl=1,keepcnt=5 /dev/ttyS2,raw,echo=0 &"
)

LOCK_FIRMWARE = "/data/busybox chattr +i "
UNLOCK_FIRMWARE = "/data/busybox chattr -i "
RUN_FTP = "/data/busybox tcpsvd -E 0.0.0.0 21 /data/busybox ftpd -w &"

# use awk because buffer
MIIO_147 = "miio_client -l 0 -o FILE_STORE -n 128 -d /data/miio"
MIIO_146 = "miio_client -l 4 -d /data/miio"
MIIO2MQTT = " | awk '/%s/{print $0;fflush()}' | mosquitto_pub -t log/miio -l &"

RE_VERSION = re.compile(r"version=([0-9._]+)")

FIRMWARE_PATHS = ("/data/firmware.bin", "/data/firmware/firmware_ota.bin")

TAR_DATA = (
    b"tar -czOC /data basic_app basic_gw conf factory miio "
    b"mijia_automation silicon_zigbee_host zigbee zigbee_gw "
    b"ble_info miioconfig.db 2>/dev/null | base64\n"
)

# default log: `tail -F /var/log/messages | grep gw3`
# publish error on app crush to mqtt: `gw3/stderr`
RUN_GW3 = "/data/gw3 -log=syslog,info 2>&1|mosquitto_pub -t gw3/stderr -s -r &"

MD5_BUSYBOX = "099137899ece96f311ac5ab554ea6fec"
MD5_GW3 = "59d94e9ba5a5336296833f0296faa985"
MD5_SOCAT = "92b77e1a93c4f4377b4b751a5390d979"


class TelnetShell(Telnet):
    def __init__(self, host: str):
        super().__init__(host, timeout=3)

        self.read_until(b"login: ", timeout=3)
        # some users have problems with \r\n symbols in login
        self.write(b"admin\n")

        raw = self.read_until(b"\r\n# ", timeout=3)
        if b"Password:" in raw:
            raise Exception("Telnet with password don't supported")

        self.ver = self.get_version()

        # adds /tmp to PATH because gw3 uses it
        self.exec("export PATH=/tmp:$PATH")

    def exec(self, command: str, as_bytes=False) -> Union[str, bytes]:
        """Run command and return it result."""
        self.write(command.encode() + b"\n")
        raw = self.read_until(b"\r\n# ", timeout=10)
        return raw if as_bytes else raw.decode()

    def check_bin(self, filename: str, md5: str, url=None) -> bool:
        """Check binary md5 and download it if needed."""
        if md5 in self.exec("md5sum /data/" + filename):
            return True
        elif url:
            self.exec(WGET.format(url, filename))
            return self.check_bin(filename, md5)
        else:
            return False

    def check_gw3(self):
        return self.check_bin("gw3", MD5_GW3)

    def run_gw3(self):
        if self.check_bin("gw3", MD5_GW3, "gw3/" + MD5_GW3):
            self.exec(RUN_GW3)

    def stop_gw3(self):
        self.exec(f"killall gw3")

    def run_zigbee_tcp(self, port=8888):
        if self.check_bin("socat", MD5_SOCAT, "bin/socat"):
            self.exec(RUN_ZIGBEE_TCP % port)

    def stop_zigbee_tcp(self):
        # stop both 8888 and 8889
        self.exec("pkill -f 'tcp-l:888'")

    def run_lumi_zigbee(self):
        self.exec("daemon_app.sh &")

    def stop_lumi_zigbee(self):
        # Z3 starts with tail on old fw and without it on new fw from 1.4.7
        self.exec("killall daemon_app.sh")
        self.exec("killall tail Lumi_Z3GatewayHost_MQTT")

    def check_firmware_lock(self) -> bool:
        """Check if firmware update locked. And create empty file if needed."""
        self.exec("mkdir -p /data/firmware")
        locked = [
            "Permission denied" in self.exec("touch " + path) for path in FIRMWARE_PATHS
        ]
        return all(locked)

    def lock_firmware(self, enable: bool):
        if self.check_bin("busybox", MD5_BUSYBOX, "bin/busybox"):
            command = LOCK_FIRMWARE if enable else UNLOCK_FIRMWARE
            for path in FIRMWARE_PATHS:
                self.exec(command + path)

    def run_ftp(self):
        if self.check_bin("busybox", MD5_BUSYBOX, "bin/busybox"):
            self.exec(RUN_FTP)

    def run_public_mosquitto(self):
        self.exec("killall mosquitto")
        time.sleep(0.5)
        self.exec("mosquitto -d")
        time.sleep(0.5)
        # fix CPU 90% full time bug
        self.exec("killall zigbee_gw")

    def run_ntpd(self):
        self.exec("ntpd -l")

    def get_running_ps(self) -> str:
        return self.exec("ps -w")

    def redirect_miio2mqtt(self, pattern: str):
        self.exec("killall daemon_miio.sh")
        self.exec("killall miio_client; pkill -f log/miio")
        time.sleep(0.5)
        cmd = MIIO_147 if self.ver >= "1.4.7_0063" else MIIO_146
        self.exec(cmd + MIIO2MQTT % pattern)
        self.exec("daemon_miio.sh &")

    def run_public_zb_console(self):
        self.stop_lumi_zigbee()

        # run Gateway with open console port (`-v` param)
        arg = " -r 'c'" if self.ver >= "1.4.7_0063" else ""

        # use `tail` because input for Z3 is required;
        # add `-l 0` to disable all output, we'll enable it later with
        # `debugprint on 1` command
        self.exec(
            "nohup tail -f /dev/null 2>&1 | "
            "nohup Lumi_Z3GatewayHost_MQTT -n 1 -b 115200 -l 0 "
            f"-p '/dev/ttyS2' -d '/data/silicon_zigbee_host/'{arg} 2>&1 | "
            "mosquitto_pub -t log/z3 -l &"
        )

        self.run_lumi_zigbee()

    def read_file(self, filename: str, as_base64=False):
        if as_base64:
            self.write(f"cat {filename} | base64\n".encode())
            self.read_until(b"\r\n", timeout=3)  # skip command
            raw = self.read_until(b"# ", timeout=10)
            try:
                # b"cat: can't open ..."
                return base64.b64decode(raw)
            except:
                return None
        else:
            self.write(f"cat {filename}\n".encode())
            self.read_until(b"\r\n", timeout=3)  # skip command
            return self.read_until(b"# ", timeout=10)[:-2]

    def tar_data(self):
        self.write(TAR_DATA)
        self.read_until(b"base64\r\n", timeout=3)  # skip command
        raw = self.read_until(b"# ", timeout=30)
        return base64.b64decode(raw)

    def run_buzzer(self):
        self.exec("kill $(ps | grep dummy:basic_gw | awk '{print $1}')")

    def stop_buzzer(self):
        self.exec("killall daemon_miio.sh")
        self.exec("killall -9 basic_gw")
        # run dummy process with same str in it
        self.exec("sh -c 'sleep 999d' dummy:basic_gw &")
        self.exec("daemon_miio.sh &")

    def get_version(self):
        raw = self.read_file("/etc/rootfs_fw_info")
        m = RE_VERSION.search(raw.decode())
        return m[1]

    def get_token(self):
        return self.read_file("/data/miio/device.token").rstrip().hex()

    def get_did(self):
        raw = self.read_file("/data/miio/device.conf").decode()
        m = re.search(r"did=(\d+)", raw)
        return m[1]

    def get_wlan_mac(self) -> str:
        raw = self.read_file("/sys/class/net/wlan0/address")

        return raw.decode().rstrip().upper()

    @property
    def mesh_group_table(self) -> str:
        if self.ver >= "1.4.7_0160":
            return "mesh_group_v3"
        elif self.ver >= "1.4.6_0043":
            return "mesh_group_v1"
        else:
            return "mesh_group"

    @property
    def mesh_device_table(self) -> str:
        if self.ver >= "1.4.7_0160":
            return "mesh_device_v3"
        else:
            return "mesh_device"

    @property
    def zigbee_db(self) -> str:
        # https://github.com/AlexxIT/XiaomiGateway3/issues/14
        # fw 1.4.6_0012 and below have one zigbee_gw.db file
        # fw 1.4.6_0030 have many json files in this folder
        return (
            "/data/zigbee_gw/*.json"
            if self.ver >= "1.4.6_0030"
            else "/data/zigbee_gw/zigbee_gw.db"
        )


NTP_DELTA = 2208988800  # 1970-01-01 00:00:00
NTP_QUERY = b"\x1b" + 47 * b"\0"


def ntp_time(host: str) -> float:
    """Return server send time"""
    try:
        sock = socket(AF_INET, SOCK_DGRAM)
        sock.settimeout(2)

        sock.sendto(NTP_QUERY, (host, 123))
        raw = sock.recv(1024)

        integ = int.from_bytes(raw[-8:-4], "big")
        fract = int.from_bytes(raw[-4:], "big")
        return integ + float(fract) / 2 ** 32 - NTP_DELTA

    except:
        return 0



import asyncio
import hashlib
import json
import logging
import random
import socket
import time
from asyncio.protocols import BaseProtocol
from asyncio.transports import DatagramTransport
from typing import Union, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

_LOGGER = logging.getLogger(__name__)

HELLO = bytes.fromhex(
    "21310020ffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
)


class BasemiIO:
    """A simple class that implements the miIO protocol."""

    device_id = None
    delta_ts = None

    def __init__(self, host: str, token: str):
        self.addr = (host, 54321)
        self.token = bytes.fromhex(token)

        key = hashlib.md5(self.token).digest()
        iv = hashlib.md5(key + self.token).digest()
        self.cipher = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=default_backend()
        )

    def _encrypt(self, plaintext: bytes):
        padder = padding.PKCS7(128).padder()
        padded_plaintext = padder.update(plaintext) + padder.finalize()

        encryptor = self.cipher.encryptor()
        return encryptor.update(padded_plaintext) + encryptor.finalize()

    def _decrypt(self, ciphertext: bytes):
        decryptor = self.cipher.decryptor()
        padded_plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        unpadder = padding.PKCS7(128).unpadder()
        return unpadder.update(padded_plaintext) + unpadder.finalize()

    def _pack_raw(self, msg_id: int, method: str, params: Union[dict, list] = None):
        # latest zero unnecessary
        payload = (
            json.dumps(
                {"id": msg_id, "method": method, "params": params or []},
                separators=(",", ":"),
            ).encode()
            + b"\x00"
        )

        data = self._encrypt(payload)

        raw = b"\x21\x31"
        raw += (32 + len(data)).to_bytes(2, "big")  # total length
        raw += b"\x00\x00\x00\x00"  # unknow
        raw += self.device_id.to_bytes(4, "big")
        raw += int(time.time() - self.delta_ts).to_bytes(4, "big")

        raw += hashlib.md5(raw + self.token + data).digest()
        raw += data

        assert len(raw) < 1024, "Exceeded message size"

        return raw

    def _unpack_raw(self, raw: bytes):
        assert raw[:2] == b"\x21\x31"
        # length = int.from_bytes(raw[2:4], 'big')
        # unknown = raw[4:8]
        # device_id = int.from_bytes(raw[8:12], 'big')
        # ts = int.from_bytes(raw[12:16], 'big')
        # checksum = raw[16:32]
        return self._decrypt(raw[32:])


class SyncmiIO(BasemiIO):
    """Synchronous miIO protocol."""

    def __init__(self, host: str, token: str, timeout: float = 3):
        super().__init__(host, token)
        self.debug = False
        self.timeout = timeout

    def ping(self, sock: socket) -> bool:
        """Returns `true` if the connection to the miio device is working. The
        token is not verified at this stage.
        """
        try:
            sock.sendto(HELLO, self.addr)
            raw = sock.recv(1024)
            if raw[:2] == b"\x21\x31":
                self.device_id = int.from_bytes(raw[8:12], "big")
                self.delta_ts = time.time() - int.from_bytes(raw[12:16], "big")
                return True
        except:
            pass
        return False

    def send(self, method: str, params: Union[dict, list] = None):
        """Send command to miIO device and get result from it. Params can be
        dict or list depend on command.
        """
        pings = 0
        for times in range(1, 4):
            try:
                # create socket every time for reset connection, because we can
                # reseive answer on previous request or request from another
                # thread
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(self.timeout)

                # need device_id for send command, can get it from ping cmd
                if self.delta_ts is None and not self.ping(sock):
                    pings += 1
                    continue

                # pack each time for new message id
                msg_id = random.randint(100000000, 999999999)
                raw_send = self._pack_raw(msg_id, method, params)
                t = time.monotonic()
                sock.sendto(raw_send, self.addr)
                # can receive more than 1024 bytes (1056 approximate maximum)
                raw_recv = sock.recv(10240)
                t = time.monotonic() - t
                data = self._unpack_raw(raw_recv).rstrip(b"\x00")

                if data == b"":
                    # mgl03 fw 1.4.6_0012 without Internet respond on miIO.info
                    # command with empty answer
                    data = {"result": ""}
                    break

                data = json.loads(data)
                # check if we received response for our cmd
                if data["id"] == msg_id:
                    break

                _LOGGER.debug(f"{self.addr[0]} | wrong ID")

            except socket.timeout:
                _LOGGER.debug(f"{self.addr[0]} | timeout {times}")
            except Exception as e:
                _LOGGER.debug(f"{self.addr[0]} | exception {e}")

            # init ping again
            self.delta_ts = None

        else:
            _LOGGER.warning(
                f"{self.addr[0]} | Device offline"
                if pings >= 2
                else f"{self.addr[0]} | Can't send {method} {params}"
            )
            return None

        if self.debug:
            _LOGGER.debug(
                f"{self.addr[0]} | Send {method} {len(raw_send)}B, "
                f"recv {len(raw_recv)}B in {t:.1f} sec and {times} try"
            )

        if "result" in data:
            return data["result"]
        else:
            _LOGGER.debug(f"{self.addr[0]} | {data}")
            return None

    def send_bulk(self, method: str, params: list):
        """Sends a command with a large number of parameters. Splits into
        multiple requests when the size of one request is exceeded.
        """
        try:
            result = []
            # Chunk of 15 is seems like the best size. Because request should
            # be lower than 1024 and response should be not more than 1056.
            # {'did':'1234567890','siid': 2,'piid': 1,'value': False,'code': 0}
            for i in range(0, len(params), 15):
                result += self.send(method, params[i : i + 15])
            return result
        except:
            return None

    def info(self) -> Union[dict, str, None]:
        """Get info about miIO device.

        Response dict - device ok, token ok
        Response empty string - device ok, token ok (mgl03 on fw 1.4.6_0012
            without cloud connection)
        Response None, device_id not None - device ok, token wrong
        Response None, device_id None - device offline
        """
        return self.send("miIO.info")


# noinspection PyMethodMayBeStatic,PyTypeChecker
class AsyncmiIO(BasemiIO, BaseProtocol):
    response = None
    sock: Optional[DatagramTransport] = None

    def datagram_received(self, data: bytes, addr):
        # hello message
        if len(data) == 32:
            if data[:2] == b"\x21\x31":
                self.device_id = int.from_bytes(data[8:12], "big")
                ts = int.from_bytes(data[12:16], "big")
                self.delta_ts = time.time() - ts
                result = True
            else:
                result = False
        else:
            result = self._unpack_raw(data)

        self.response.set_result(result)

    def error_received(self, exc):
        print("Error received:", exc)

    def connection_lost(self, exc):
        print("Connection closed")
        self.sock = None

    async def send_raw(self, data: bytes):
        loop = asyncio.get_running_loop()
        if not self.sock:
            self.sock, _ = await loop.create_datagram_endpoint(
                lambda: self, remote_addr=self.addr
            )

        self.response = loop.create_future()
        # this method does not block
        self.sock.sendto(data)
        return await self.response

    async def ping(self):
        return await self.send_raw(HELLO)

    async def send(self, method: str, params: Union[dict, list] = None):
        """Send command to miIO device and get result from it. Params can be
        dict or list depend on command.
        """
        if not self.device_id and not await self.ping():
            return None

        try:
            raw = self._pack_raw(method, params)
            data = await self.send_raw(raw)
            return json.loads(data.rstrip(b"\x00"))["result"]
        except Exception as e:
            _LOGGER.warning(f"Can't send: {e}")
            return None

    async def send_bulk(self, method: str, params: list):
        """Sends a command with a large number of parameters. Splits into
        multiple requests when the size of one request is exceeded.
        """
        try:
            result = []
            for i in range(0, len(params), 15):
                result += await self.send(method, params[i : i + 15])
            return result
        except:
            return None

    async def info(self):
        """Get info about miIO device."""
        return await self.send("miIO.info")



from __future__ import annotations

import asyncio
import functools

from .. import fhem, utils
from ..generic import FhemModule

from pyvaillant import ClientAuth, VaillantThermostatData


class erelax_vaillant(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        unknown_key = "thisisnotmagic012345678901234567"
        self.unknown_code = utils.decrypt_string(
            (
                "eJwVwdsOgiAAANAvalNLrEcJM5chh"
                "RvOF6dZaBcCLBG/vnUOD/8gr1RNJ"
                "iNsgTOZrebk5njuM7Es\nlo4SBja"
                "S+yoZ+YHh61CIsF/qvE6Fsmr30sh"
                "FXeCt+WNoiAZ9R++EGRmXZ5pH+zh"
                "tWbjxzZa+JwjQ\nFGFEmvJ7vFgI3"
                "NYLPgs869OIiU9/UCcvrg==\n"
            ),
            unknown_key,
        )

        attr_config = {
            "update_interval": {
                "default": 300,
                "format": "int",
                "help": "Change interval, default is 300.",
            }
        }
        self.set_attr_config(attr_config)

        set_config = {
            "away": {},
            "home": {},
            "desiredTemperature": {
                "args": ["temperature"],
                "options": "slider,7,0.5,30",
            },
            "system_mode": {
                "args": ["mode"],
                "argsh": ["mode"],
                "options": "winter,summer,frostguard",
            },
        }
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define erelax_vaillant fhempy USERNAME PASSWORD [STATION]"
        await fhem.readingsSingleUpdate(hash, "state", "connecting", 1)
        self.create_async_task(self.do_update_loop(args[3], args[4]))

    async def set_away(self, hash, params):
        self.create_async_task(self.v_station.get_modules()[0].set_mode("away"))

    async def set_home(self, hash, params):
        self.create_async_task(self.v_station.get_modules()[0].disable_mode("away"))

    async def set_desiredTemperature(self, hash, params):
        self.create_async_task(
            self.v_station.get_modules()[0].set_setpoint_temp(params["temperature"])
        )

    async def set_system_mode(self, hash, params):
        self.create_async_task(
            self.v_station.get_modules()[0].set_system_mode(params["mode"])
        )

    async def do_update_loop(self, username, password):
        await self.do_login(username, password)

        self.v_station = VaillantStation(self.logger, self.v_auth)
        await self.v_station.init_device()
        await self.update_readings()

        self.v_station.register_update_readings(self.update_readings)

        while True:
            await asyncio.sleep(self._attr_update_interval)
            try:
                await self.v_station.update()
            except Exception as ex:
                self.logger.exception(ex)

    async def do_login(self, username, password):
        for i in range(3):
            try:
                self.v_auth = await utils.run_blocking(
                    functools.partial(
                        ClientAuth,
                        "na_client_android_vaillant",
                        self.unknown_code,
                        username,
                        password,
                        (
                            "read_station read_camera access_camera read_thermostat "
                            "write_thermostat read_presence access_presence"
                        ),
                        "1.0.4.0",
                        "vaillant",
                    )
                )
                await asyncio.sleep(5)
                return
            except Exception as ex:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "Failed to login to netatmo API", 1
                )
                self.logger.exception(ex)

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        v_station = self.v_station
        await fhem.readingsBulkUpdate(self.hash, "station_name", v_station.station_name)
        await fhem.readingsBulkUpdate(self.hash, "system_mode", v_station.system_mode)
        await fhem.readingsBulkUpdate(self.hash, "station_mac", v_station.mac)
        await fhem.readingsBulkUpdate(self.hash, "station_type", v_station.model_type)
        await fhem.readingsBulkUpdate(
            self.hash, "place_altitude", v_station.place_altitude
        )
        await fhem.readingsBulkUpdate(self.hash, "place_city", v_station.place_city)
        await fhem.readingsBulkUpdate(
            self.hash, "place_continent", v_station.place_continent
        )
        await fhem.readingsBulkUpdate(
            self.hash, "place_country_name", v_station.place_country_name
        )
        await fhem.readingsBulkUpdate(self.hash, "place_street", v_station.place_street)
        await fhem.readingsBulkUpdate(
            self.hash, "place_timezone", v_station.place_timezone
        )
        await fhem.readingsBulkUpdate(
            self.hash, "place_location", v_station.place_location
        )
        await fhem.readingsBulkUpdate(self.hash, "dhw", v_station.dhw)
        await fhem.readingsBulkUpdate(self.hash, "station_firmware", v_station.firmware)
        await fhem.readingsBulkUpdate(self.hash, "oem_serial", v_station.oem_serial)
        await fhem.readingsBulkUpdate(
            self.hash, "hot_water_anticipating", v_station.hot_water_anticipating
        )
        await fhem.readingsBulkUpdate(self.hash, "wifi_status", v_station.wifi_status)
        await fhem.readingsBulkUpdate(self.hash, "station_firmware", v_station.firmware)
        await fhem.readingsBulkUpdate(
            self.hash, "boiler_oem_serial", v_station.boiler_oem_serial
        )
        await fhem.readingsBulkUpdate(
            self.hash, "outdoor_temperature", v_station.outdoor_temperatur
        )
        await fhem.readingsBulkUpdate(
            self.hash, "refill_water", v_station.ebus_refill_water
        )
        # currently only one thermostat (module) is supported
        v_module = self.v_station.get_modules()[0]
        await fhem.readingsBulkUpdate(self.hash, "thermostat_name", v_module.name)
        await fhem.readingsBulkUpdate(self.hash, "temperature", v_module.temperature)
        await fhem.readingsBulkUpdate(self.hash, "thermostat_mac", v_module.mac)
        await fhem.readingsBulkUpdate(
            self.hash, "thermostat_firmware", v_module.firmware
        )
        await fhem.readingsBulkUpdate(self.hash, "rf_status", v_module.rf_status)
        await fhem.readingsBulkUpdate(
            self.hash, "battery_percent", v_module.battery_percent
        )
        await fhem.readingsBulkUpdate(
            self.hash, "est_setpoint_temp", v_module.est_setpoint_temp
        )
        await fhem.readingsBulkUpdate(
            self.hash, "desiredTemperature", v_module.setpoint_temp
        )
        await fhem.readingsBulkUpdate(
            self.hash, "away", v_module.setpoint_away["setpoint_activate"]
        )
        if v_module.setpoint_away["setpoint_activate"]:
            await fhem.readingsBulkUpdate(
                self.hash, "away_endtime", v_module.setpoint_away["setpoint_endtime"]
            )
        else:
            await fhem.readingsBulkUpdate(self.hash, "away_endtime", "0")
        await fhem.readingsBulkUpdate(
            self.hash, "manual", v_module.setpoint_manual["setpoint_activate"]
        )
        if v_module.setpoint_manual["setpoint_activate"]:
            await fhem.readingsBulkUpdate(
                self.hash,
                "manual_endtime",
                v_module.setpoint_manual["setpoint_endtime"],
            )
        else:
            await fhem.readingsBulkUpdate(
                self.hash,
                "manual_endtime",
                "0",
            )
        await fhem.readingsEndUpdate(self.hash, 1)


class VaillantStation:
    def __init__(self, logger, auth: ClientAuth, name: str = None) -> VaillantStation:
        self.auth = auth
        self.name = name
        self.logger = logger
        self.raw_device = None
        self.modules = []
        self._update_reading_function = None

    async def init_device(self):
        try:
            self._vaillant = await utils.run_blocking(
                functools.partial(VaillantThermostatData, self.auth)
            )
            self.raw_device = self.get_station_by_name(self.name)
            for raw_module in self.raw_device["modules"]:
                self.modules.append(VaillantModule(self.logger, self, raw_module))
        except Exception as ex:
            self.logger.exception(ex)

    def register_update_readings(self, update_reading_function):
        self._update_reading_function = update_reading_function

    def update_token(self):
        self._vaillant.token = self.auth.accessToken

    async def update(self):
        await utils.run_blocking(functools.partial(self.update_token))
        await utils.run_blocking(functools.partial(self._vaillant.update))
        self.raw_device = self.get_station_by_name(self.name)
        for module in self.modules:
            module.update(self.raw_device)
        if self._update_reading_function is not None:
            await self._update_reading_function()

    def get_station_by_name(self, name: str = None):
        if name is None:
            return self._vaillant.devList[0]
        for key, value in self._vaillant.devList.items():
            if value["station_name"] == name:
                return self._vaillant.devList[key]

    def get_modules(self) -> list[VaillantModule]:
        return self.modules

    @property
    def vaillant(self) -> VaillantThermostatData:
        return self._vaillant

    @vaillant.setter
    def vaillant(self, vaillant: VaillantThermostatData):
        self._vaillant = vaillant

    @property
    def station_name(self):
        if self.raw_device is not None:
            return self.raw_device["station_name"]
        return ""

    @property
    def system_mode(self):
        if self.raw_device is not None:
            return self.raw_device["system_mode"]
        return ""

    @property
    def mac(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["_id"]

    @property
    def model_type(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["type"]

    @property
    def place_altitude(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["altitude"]

    @property
    def place_city(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["city"]

    @property
    def place_continent(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["continent"]

    @property
    def place_country_name(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["country_name"]

    @property
    def place_street(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["street"]

    @property
    def place_timezone(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["place"]["timezone"]

    @property
    def place_location(self):
        if self.raw_device is None:
            return ""
        return (
            str(self.raw_device["place"]["location"][1])
            + ","
            + str(self.raw_device["place"]["location"][0])
        )

    @property
    def dhw(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["dhw"]

    @property
    def firmware(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["firmware"]

    @property
    def oem_serial(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["oem_serial"]

    @property
    def hot_water_anticipating(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["hot_water_anticipating"]

    @property
    def wifi_status(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["wifi_status"]

    @property
    def boiler_oem_serial(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["boiler_oem_serial"]

    @property
    def outdoor_temperatur(self):
        if self.raw_device is None:
            return 0.0
        return self.raw_device["outdoor_temperature"]["te"]

    @property
    def ebus_refill_water(self):
        if self.raw_device is None:
            return False
        return self.raw_device["ebus_status"]["refill_water"]

    @property
    def setpoint_hwb(self):
        if self.raw_device is None:
            return ""
        return self.raw_device["setpoint_hwb"]


class VaillantModule:
    def __init__(self, logger, station: VaillantStation, raw_module):
        self.logger = logger
        self.station = station
        self.module_name = raw_module["module_name"]
        self.raw_module = raw_module

    def update(self, data):
        for module in data["modules"]:
            if module["module_name"] == self.module_name:
                self.raw_module = module

    @property
    def name(self) -> str:
        if self.raw_module is None:
            return ""
        return self.raw_module["module_name"]

    @property
    def firmware(self) -> str:
        if self.raw_module is None:
            return ""
        return self.raw_module["firmware"]

    @property
    def rf_status(self) -> str:
        if self.raw_module is None:
            return ""
        return self.raw_module["rf_status"]

    @property
    def battery_percent(self) -> str:
        if self.raw_module is None:
            return ""
        return self.raw_module["battery_percent"]

    @property
    def setpoint_away(self):
        if self.raw_module is None:
            return {"setpoint_activate": False}
        return self.raw_module["setpoint_away"]

    @property
    def setpoint_manual(self):
        if self.raw_module is None:
            return {"setpoint_activate": False}
        return self.raw_module["setpoint_manual"]

    @property
    def temperature(self) -> float:
        if self.raw_module is None:
            return 0.0
        return self.raw_module["measured"]["temperature"]

    @property
    def mac(self):
        if self.raw_module is None:
            return ""
        return self.raw_module["_id"]

    @property
    def setpoint_temp(self):
        if self.raw_module is None:
            return 0.0
        return self.raw_module["measured"]["setpoint_temp"]

    async def set_setpoint_temp(self, temp):
        try:
            await utils.run_blocking_task(
                functools.partial(self.station.vaillant.activate, "manual", temp)
            )
            # wait for update at netatmo
            await asyncio.sleep(1)
            await self.station.update()
        except Exception as ex:
            self.logger.exception(ex)

    async def set_mode(self, mode):
        try:
            # this is not handled but a required parameter within the function
            settemp = 15
            await utils.run_blocking_task(
                functools.partial(self.station.vaillant.activate, mode, settemp)
            )
            # wait for update at netatmo
            await asyncio.sleep(1)
            await self.station.update()
        except Exception as ex:
            self.logger.exception(ex)

    async def disable_mode(self, mode):
        try:
            await utils.run_blocking_task(
                functools.partial(self.station.vaillant.disable, mode)
            )
            # wait for update at netatmo
            await asyncio.sleep(1)
            await self.station.update()
        except Exception as ex:
            self.logger.exception(ex)

    async def set_system_mode(self, mode):
        try:
            await utils.run_blocking_task(
                functools.partial(self.station.vaillant.setSystemMode, mode)
            )
            # wait for update at netatmo
            await asyncio.sleep(1)
            await self.station.update()
        except Exception as ex:
            self.logger.exception(ex)

    @property
    def est_setpoint_temp(self):
        if self.raw_module is None:
            return 0.0
        return self.raw_module["measured"]["est_setpoint_temp"]



import json
import logging
import random
import re
import string
import time
import functools
from typing import List, Optional

from fhempy.lib import utils

import requests

from .mini_miio import SyncmiIO
from .shell import TelnetShell
from .xiaomi_cloud import MiCloud

DOMAIN = "xiaomi_gateway3"

_LOGGER = logging.getLogger(__name__)


# new miio adds colors to logs
RE_JSON1 = re.compile(b"msg:(.+) length:([0-9]+) bytes")
RE_JSON2 = re.compile(b"{.+}")


def extract_jsons(raw) -> List[bytes]:
    """There can be multiple concatenated json on one line. And sometimes the
    length does not match the message."""
    m = RE_JSON1.search(raw)
    if m:
        length = int(m[2])
        raw = m[1][:length]
    else:
        m = RE_JSON2.search(raw)
        raw = m[0]
    return raw.replace(b"}{", b"}\n{").split(b"\n")


def migrate_options(data):
    data = dict(data)
    options = {k: data.pop(k) for k in ("ble", "zha") if k in data}
    return {"data": data, "options": options}


def check_mgl03(host: str, token: str, telnet_cmd: Optional[str]) -> Optional[str]:
    try:
        # 1. try connect with telnet (custom firmware)?
        shell = TelnetShell(host)
        # 1.1. check token with telnet
        return None if shell.get_token() == token else "wrong_token"
    except:
        if not telnet_cmd:
            return "cant_connect"

    # 2. try connect with miio
    miio = SyncmiIO(host, token)
    info = miio.info()
    # fw 1.4.6_0012 without cloud will respond with a blank string reply
    if info is None:
        # if device_id not None - device works but not answer on commands
        return "wrong_token" if miio.device_id else "cant_connect"

    # 3. check if right model
    if info and info["model"] != "lumi.gateway.mgl03":
        return "wrong_model"

    raw = json.loads(telnet_cmd)
    # fw 1.4.6_0043+ won't answer on cmd without cloud, so don't check answer
    miio.send(raw["method"], raw.get("params"))

    # waiting for telnet to start
    time.sleep(1)

    try:
        # 4. check if telnet command helps
        TelnetShell(host)
    except:
        return "wrong_telnet"


def get_lan_key(host: str, token: str):
    device = SyncmiIO(host, token)
    resp = device.send("get_lumi_dpf_aes_key")
    if resp is None:
        return "Can't connect to gateway"
    if len(resp[0]) == 16:
        return resp[0]
    key = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(16)
    )
    resp = device.send("set_lumi_dpf_aes_key", [key])
    if resp[0] == "ok":
        return key
    return "Can't update gateway key"


async def get_room_mapping(cloud: MiCloud, host: str, token: str):
    try:
        device = SyncmiIO(host, token)
        local_rooms = device.send("get_room_mapping")
        cloud_rooms = await cloud.get_rooms()
        result = ""
        for local_id, cloud_id in local_rooms:
            cloud_name = next(
                (p["name"] for p in cloud_rooms if p["id"] == cloud_id), "-"
            )
            result += f"\n- {local_id}: {cloud_name}"
        return result

    except:
        return "Can't get from cloud"


async def get_bindkey(cloud: MiCloud, did: str):
    bindkey = await cloud.get_bindkey(did)
    if bindkey is None:
        return "Can't get from cloud"
    if bindkey.endswith("FFFFFFFF"):
        return "Not needed"
    return bindkey


def reverse_mac(s: str):
    return f"{s[10:]}{s[8:10]}{s[6:8]}{s[4:6]}{s[2:4]}{s[:2]}"


EZSP_URLS = {
    7: "https://master.dl.sourceforge.net/project/mgl03/zigbee/"
    "ncp-uart-sw_mgl03_6_6_2_stock.gbl?viasf=1",
    8: "https://master.dl.sourceforge.net/project/mgl03/zigbee/"
    "ncp-uart-sw_mgl03_6_7_8_z2m.gbl?viasf=1",
}


def _update_zigbee_firmware(host: str, ezsp_version: int):
    shell = TelnetShell(host)

    # stop all utilities without checking if they are running
    shell.stop_lumi_zigbee()
    shell.stop_zigbee_tcp()
    # flash on another port because running ZHA or z2m can breake process
    shell.run_zigbee_tcp(port=8889)
    time.sleep(0.5)

    _LOGGER.debug(f"Try update EZSP to version {ezsp_version}")

    from .elelabs_ezsp_utility import ElelabsUtilities

    config = type(
        "", (), {"port": (host, 8889), "baudrate": 115200, "dlevel": _LOGGER.level}
    )
    utils = ElelabsUtilities(config, _LOGGER)

    # check current ezsp version
    resp = utils.probe()
    _LOGGER.debug(f"EZSP before flash: {resp}")
    if resp[0] == 0 and resp[1] == ezsp_version:
        return True

    url = EZSP_URLS[ezsp_version]
    r = requests.get(url)

    resp = utils.flash(r.content)
    _LOGGER.debug(f"EZSP after flash: {resp}")
    return resp[0] == 0 and resp[1] == ezsp_version


async def update_zigbee_firmware(host: str, ezsp_version: int):
    return await utils.run_blocking(
        functools.partial(_update_zigbee_firmware, host, ezsp_version)
    )
